import json
import logging
import logging.config
import requests
from trainify.utils import send_to_backend


class BackendHandler(logging.Handler, object):
    """
    后端实时输出handler
    """

    def __init__(self, channel, other_attr=None):
        logging.Handler.__init__(self)
        self.channel = channel
        # print('初始化自定义日志处理器：', channel)
        # print('其它属性值：', other_attr)

    def emit(self, record):
        """
        emit函数为自定义handler类时必重写的函数，这里可以根据需要对日志消息做一些处理，比如发送日志到服务器
        发出记录(Emit a record)
        """
        try:
            msg = self.format(record)
            content = {
                "type": "msg",
                "data": {
                    "content": msg
                }
            }
            send_to_backend(self.channel, content)
        except Exception:
            self.handleError(record)



"""
.. module:: language
   :synopsis: Represents formulas and propositional logics.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

import sys

from .. import language as BooleanLogics

from ..language import LNot
from ..language import get_alphabet


class Formula(BooleanLogics.Formula):
    ''' A class to represent propositional formulas.

    Formulas are represented as nodes in labelled trees: leaves are terminal
    symbols (e.g., atomic propositions and Boolean values), while internal
    nodes correspond to operators and quantifiers. The ariety of internal nodes
    depends on the kind of operator or quantifier must be represented.
    For instance, the arity of a node representing the formula
    :math:`not (p \lor True)` is one because the formula has exclusively one
    sub-formula, i.e., :math:`p \lor True`. On the contrary, this last formula
    has two sub-formulas, i.e., :math:`p` and  :math:`True`, thus, the node
    representing it has two sons.
    '''

    __desc__ = 'propositional formula'

    def __init__(self, *phi):
        self.wrap_subformulas(phi, sys.modules[self.__module__].Formula)

    def wrap_subformulas(self, subformulas, FormulaClass):
        ''' Replaces subformulas of the current object

        This method replaces the subformulas of the current object by
        using the :class:`FormulaClass` objects representing the
        elements of :param subformulas:.

        :param self: this object
        :type self: PL.Formula
        :param subformulas: the set of objects to be used as model for the
            replacement
        :type subformulas: Container
        :param FormulaClass: the final class of the new subformulas
        :type FormulaClass: class
        '''

        def err_msg(phi):
            return ('expected a {}, '.format(FormulaClass.__desc__) +
                    'got the {} {}'.format(phi.__desc__, phi))

        Lang = sys.modules[self.__module__]

        self._subformula = []
        self.height = 0

        for phi in subformulas:
            if isinstance(phi, bool):
                self._subformula.append(Lang.Bool(phi))
            else:
                if isinstance(phi, str):
                    self._subformula.append(Lang.AtomicProposition(phi))
                else:
                    if not isinstance(phi, FormulaClass):
                        if (isinstance(phi, Lang.Formula) or
                                not isinstance(phi, Formula)):

                            raise TypeError(err_msg(phi))

                        phi = phi.cast_to(Lang)

                        if not isinstance(phi, FormulaClass):
                            raise TypeError(err_msg(phi))

                    self._subformula.append(phi)
                    self.height = max(self.height, phi.height+1)

    def cast_to(self, Lang):
        ''' Casts the current object in a formula of a different class.

        :param self: this formula
        :type self: Formula
        :param Lang: a class representing a language
        :type self: class
        :returns: a syntactically equivalent formula in language represented by
           :class:`Lang`
        :rtype: Lang
        '''
        if isinstance(self, Bool):
            return Lang.Bool(bool(self._value))
        if isinstance(self, AtomicProposition):
            return Lang.AtomicProposition(str(self.name))

        symbol_name = self.__class__.__name__
        if symbol_name not in Lang.alphabet:
            raise TypeError('{} is not in the alphabet '.format(symbol_name) +
                            'of {}, thus {} is '.format(Lang.__name__, self) +
                            'not a {} formula'.format(Lang.__name__))

        subformulas = []
        for subformula in self.subformulas():
            subformulas.append(Formula.cast_to(subformula, Lang))

        return Lang.alphabet[symbol_name](*subformulas)


class AtomicProposition(Formula, BooleanLogics.AlphabeticSymbol):
    '''
    The class representing atomic propositionic propositions.

    '''
    def __init__(self, name):
        ''' Initialize a atomic proposition.

        This method builds a atomic propositionic proposition.

        :param name: the name of the atomic proposition.
        :type name: str
        '''
        if not isinstance(name, str):
            raise TypeError('name = \'{}\' must be '.format(name) +
                            'a {} object, but it is '.format(str) +
                            '{} object'.format(name.__class__))
        self.name = str(name)
        self.height = 0

    def clone(self):
        ''' Clones an atomic proposition

        :returns: a clone of the current atomic proposition
        :rtype: PL.AtomicProposition
        '''
        return self.__class__(str(self.name))

    def subformula(self, i):
        ''' Returns the :math:`i`-th subformula.

        :param i: the index of the subformula to be returned
        :type i: Integer
        :raise TypeError: atomic propositions have not subformulas
        '''
        raise TypeError('AtomicPropositions have not subformulas.')

    def subformulas(self):
        ''' Returns the list of all the subformulas.

        :returns: returns the empty list of the subformulas of the current
            formula
        :rtype: list
        '''
        return []

    def __str__(self):
        return '{}'.format(self.name)


class Bool(BooleanLogics.Bool, AtomicProposition):
    '''
    The class of Boolean atomic propositions.

    '''
    pass


class LogicOperator(Formula, BooleanLogics.LogicOperator):
    '''
    A class to represent logic operator such as :math:`\land` or :math:`\lor`.

    '''
    pass


class Not(LogicOperator, BooleanLogics.Not):
    '''
    Represents logic negation.

    '''
    pass


class Or(LogicOperator,  BooleanLogics.Or):
    '''
    Represents logic non-exclusive disjunction.

    '''
    pass


class And(LogicOperator, BooleanLogics.And):
    '''
    Represents logic conjunction.

    '''
    pass


class Imply(LogicOperator, BooleanLogics.Imply):
    '''
    Represents logic implication.

    '''
    pass


def get_symbols(alphabet):
    symbols = list()

    for name, c in alphabet.items():
        if name == 'Bool':
            symbols.extend(Bool.symbols.values())
        else:
            if name != 'AtomicProposition':
                symbols.extend(c.symbols)

    return symbols


alphabet = get_alphabet(__name__)
symbols = get_symbols(alphabet)



"""
.. module:: CTLS.language
   :synopsis: Represents the CTL* language.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from trainify.validator.pyModelChecking import PL

from ..language import LNot
from ..language import get_alphabet
from trainify.validator.pyModelChecking.PL import get_symbols

from ..language import AlphabeticSymbol

import sys


class Formula(PL.Formula):
    ''' A class to represent CTL* formulas.

    Formulas are represented as nodes in labelled trees: leaves are
    terminal symbols (e.g., atomic propositions and Boolean values),
    while internal nodes correspond to operators and quantifiers.
    The ariety of internal nodes depends on the kind of operator or
    quantifier must be represented. For instance, the arity of a node
    representing the formula :math:`not (A(p U q) \lor True)` is one because
    the formula has exclusively one sub-formula, i.e.,
    :math:`A(p U q) \lor True`. On the contrary, this last formula has
    two sub-formulas, i.e., :math:`A(p U q)` and  :math:`True`, thus, the node
    representing it has two sons.
    '''

    __desc__ = 'CTL* formula'

    def __init__(self, *phi):
        self.wrap_subformulas(phi, sys.modules[self.__module__].Formula)

    def get_equivalent_non_fair_formula(self, fairAP):
        fair_sfs = [sf.get_equivalent_non_fair_formula(fairAP)
                    for sf in self._subformula]

        return self.__class__(*fair_sfs)

    def is_a_state_formula(self):
        ''' Returns True if and only if the object represents a state formula.

        This method should return True if and only if the current object
        represents a state formula. Since this is a general class meant to
        represent both state and path formulas, an implementation for this
        method cannot be provided. Thus, any call to it raise a
        :class:`RuntimeError`.

        :param self: this formula
        :type self: CTLS.Formula
        :raise RuntimeError: this method cannot be implemented by this general
            class
        '''
        raise RuntimeError('{}.is_a_state_formula() '.format(self.__class__) +
                           'not implemented yet')


class PathFormula(Formula):
    '''
    A class representing CTL* path formulas.

    '''

    __desc__ = 'CTL* path formula'

    def is_a_state_formula(self):
        ''' Returns True if and only if the object has type :class:`StateFormula`.

        This method returns True if and only if the current object has
        type :class:`CTLS.StateFormula`. Since this is a method of the
        class :class:`CTLS.PathFormula`, it always returns False.

        :param self: this formula
        :type self: CTLS.PathFormula
        :returns: False
        :rtype: bool
        '''
        return False

    def __init__(self, *phi):
        self.wrap_subformulas(phi, sys.modules[self.__module__].Formula)


class StateFormula(PathFormula):
    '''
    A class representing CTL* state formulas.

    '''

    __desc__ = 'CTL* state formula'

    def is_a_state_formula(self):
        ''' Returns True if and only if the object has type :class:`StateFormula`.

        This method returns True if and only if the current object has
        type :class:`CTLS.StateFormula`. Since this is a method of the
        class :class:`CTLS.StateFormula`, it always returns True.

        :param self: this formula
        :type self: CTLS.StateFormula
        :returns: True
        :rtype: bool
        '''
        return True

    def __init__(self, *phi):
        self.wrap_subformulas(phi, sys.modules[self.__module__].Formula)


class AtomicProposition(PL.AtomicProposition, StateFormula):
    '''
    The class representing atomic propositionic propositions.

    '''

    def __init__(self, name):
        ''' Initialize a CTL* atomic proposition.

        This method builds a CTL* atomic propositionic proposition.

        :param name: the name of the atomic proposition.
        :type name: str
        '''
        super(AtomicProposition, self).__init__(name)

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula. Since this is a method of the class
        :class:`CTLS.AtomicProposition`, a clone of the current object is
        always returned.

        :returns: a clone of the current atomic proposition.
        :rtype: CTLS.AtomicProposition
        '''
        return self.clone()

    def get_equivalent_non_fair_formula(self, fairAP):
        Lang = sys.modules[self.__module__]

        return Lang.And(self, fairAP)


class Bool(PL.Bool, AtomicProposition):
    '''
    The class of Boolean atomic propositions.

    '''

    def __init__(self, value):
        ''' Initialize a Boolean atomic proposition.

        This method builds either a True or a False atomic proposition.

        :param value: the boolean atomic proposition.
        :type value: bool

        '''
        super(Bool, self).__init__(value)


class TemporalOperator(PathFormula):
    '''
    A class to represent temporal operators such as :math:`R` or :math:`X`.

    '''

    def __str__(self):
        if len(self._subformula) == 1:
            return '{}({})'.format(self.__class__.symbols[0],
                                   self._subformula[0])
        else:
            sep = ' {} '.format(self.__class__.symbols[0])
            return '({})'.format(sep.join([str(f) for f in self._subformula]))


class PathQuantifier(StateFormula):
    '''
    A class to represent the path quantifiers :math:`A` or :math:`E`.

    '''

    def __init__(self, phi):
        self.wrap_subformulas([phi], Formula)

    def __str__(self):
        return '{}({})'.format(self.__class__.symbols[0], self._subformula[0])


class LogicOperator(Formula, PL.LogicOperator):
    '''
    A class to represent logic operator such as :math:`\land` or :math:`\lor`.

    '''

    def is_a_state_formula(self):
        ''' Returns True if and only if the object represents a state formula.

        This method returns True if and only if the current object represents
        a state formula.

        :param self: this formula
        :type self: CTLS.LogicOperator
        :returns: True if and only if the object represents a state formula.
        :rtype: bool
        '''
        for sf in self.subformulas():
            if not sf.is_a_state_formula():
                return False

        return True


class Not(LogicOperator, PL.Not):

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.Not
        '''
        subformula = self.subformula(0).get_equivalent_restricted_formula()

        return LNot(subformula)

    def is_a_state_formula(self):
        ''' Returns True if and only if the object represents a state formula.

        This method returns True if and only if the current object represents
        a state formula.

        :param self: this formula
        :type self: CTLS.Not
        :returns: True if and only if this formula is a state formula.
        :rtype: bool
        '''
        return self.subformula(0).is_a_state_formula()


class A(PathQuantifier, AlphabeticSymbol):
    '''
    A class representing CTL* A-formulas.

    '''

    symbols = ['A']

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.Not
        '''
        subformula = self.subformula(0).get_equivalent_restricted_formula()

        Lang = sys.modules[self.__module__]
        return Lang.Not(Lang.E(LNot(subformula)))

    def get_equivalent_non_fair_formula(self, fairAP):
        Lang = sys.modules[self.__module__]

        sf = self.subformula(0).get_equivalent_non_fair_formula(fairAP)

        return self.__class__(LNot(Lang.And(LNot(sf), fairAP)))


class E(PathQuantifier, AlphabeticSymbol):
    '''
    A class representing CTL* A-formulas.

    '''

    symbols = ['E']

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.E
        '''
        subformula = self.subformula(0).get_equivalent_restricted_formula()

        Lang = sys.modules[self.__module__]
        return Lang.E(subformula)

    def get_equivalent_non_fair_formula(self, fairAP):
        fair_sf = self.subformula(0).get_equivalent_non_fair_formula(fairAP)

        Lang = sys.modules[self.__module__]
        return Lang.E(Lang.And(fairAP, fair_sf))


class X(TemporalOperator, AlphabeticSymbol):
    symbols = ['X']

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and
                  is equivalent this formula
        :rtype: CTLS.X
        '''
        subformula = self.subformula(0).get_equivalent_restricted_formula()

        Lang = sys.modules[self.__module__]
        return Lang.X(subformula)


class F(TemporalOperator, AlphabeticSymbol):
    symbols = ['F']

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.U
        '''
        subformula = self.subformula(0).get_equivalent_restricted_formula()

        Lang = sys.modules[self.__module__]
        return Lang.U(True, subformula)


class G(TemporalOperator, AlphabeticSymbol):
    symbols = ['G']

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.Not
        '''
        subformula = self.subformula(0).get_equivalent_restricted_formula()

        Lang = sys.modules[self.__module__]
        return Lang.Not(Lang.U(True, LNot(subformula)))


class Or(LogicOperator, PL.Or):
    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.Or
        '''
        subformulas = [p.get_equivalent_restricted_formula()
                       for p in self._subformula]

        Lang = sys.modules[self.__module__]
        return Lang.Or(*subformulas)


class And(LogicOperator, PL.And):
    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.Not
        '''
        subformulas = []
        for p in self._subformula:
            equi_p = p.get_equivalent_restricted_formula()
            subformulas.append(LNot(equi_p))

        Lang = sys.modules[self.__module__]
        return Lang.Not(Lang.Or(*subformulas))


class Imply(LogicOperator, PL.Imply):
    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.Not
        '''
        equiv_sf0 = self.subformula(0).get_equivalent_restricted_formula()

        Lang = sys.modules[self.__module__]
        return Lang.Or(LNot(equiv_sf0),
                       self.subformula(1).get_equivalent_restricted_formula())


class U(TemporalOperator, AlphabeticSymbol):
    symbols = ['U']

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.U
        '''
        subformulas = []
        for p in self._subformula:
            subformulas.append(p.get_equivalent_restricted_formula())

        Lang = sys.modules[self.__module__]
        return Lang.U(*subformulas)


class R(TemporalOperator, AlphabeticSymbol):
    symbols = ['R']

    def get_equivalent_restricted_formula(self):
        ''' Return an equivalent formula in the restricted syntax.

        This method returns a formula that avoids :math:`A`, :math:`F`,
        :math:`R`, :math:`\land` and :math:`\\rightarrow` and is equivalent to
        this formula.

        :returns: a formula that avoids :math:`A`, :math:`F`, :math:`R`,
                  :math:`\land` and :math:`\\rightarrow` and is equivalent to
                  this formula
        :rtype: CTLS.Not
        '''
        subformulas = []
        for p in self._subformula:
            equi_p = p.get_equivalent_restricted_formula()
            subformulas.append(LNot(equi_p))

        Lang = sys.modules[self.__module__]
        return Lang.Not(Lang.U(*subformulas))


alphabet = get_alphabet(__name__)
symbols = get_symbols(alphabet)



"""
.. module:: LTL.language
   :synopsis: Represents the LTL language.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

import sys

from ..language import LNot
from ..language import get_alphabet, get_symbols

import trainify.validator.pyModelChecking.CTLS

CTLS = sys.modules['trainify.validator.pyModelChecking.CTLS']


class Formula(CTLS.Formula):
    '''
    A class representing LTL formulas.

    '''

    __desc__ = 'LTL formula'


class PathFormula(Formula, CTLS.PathFormula):
    '''
    A class representing LTL path formulas.

    '''

    __desc__ = 'LTL path formula'

    def __init__(self, *phi):
        self.wrap_subformulas(phi, PathFormula)


class X(PathFormula, CTLS.X):
    '''
    A class representing LTL X-formulas.

    '''

    pass


class F(PathFormula, CTLS.F):
    '''
    A class representing LTL F-formulas.

    '''

    pass


class G(PathFormula, CTLS.G):
    '''
    A class representing LTL G-formulas.

    '''

    pass


class U(PathFormula, CTLS.U):
    '''
    A class representing LTL U-formulas.

    '''

    pass


class R(PathFormula, CTLS.R):
    '''
    A class representing LTL R-formulas.

    '''

    pass


class AtomicProposition(CTLS.AtomicProposition, PathFormula):
    '''
    A class representing LTL atomic propositions.

    '''

    pass


class Bool(CTLS.Bool, PathFormula):
    '''
    A class representing LTL Boolean atomic propositions.

    '''

    pass


class Not(PathFormula, CTLS.Not):
    '''
    A class representing LTL negations.

    '''

    pass


class Or(PathFormula, CTLS.Or):
    '''
    A class representing LTL disjunctions.

    '''

    pass


class And(PathFormula, CTLS.And):
    '''
    A class representing LTL conjunctions.

    '''

    pass


class Imply(PathFormula, CTLS.Imply):
    '''
    A class representing LTL implications.

    '''

    pass


class StateFormula(Formula, CTLS.StateFormula):
    '''
    A class representing LTL state formulas.

    '''

    __desc__ = 'LTL state formula'

    def __init__(self, *phi):
        self.wrap_subformulas(phi, PathFormula)


class A(StateFormula, CTLS.A):
    '''
    A class representing LTL A-formulas.

    '''
    pass


alphabet = get_alphabet(__name__)
symbols = get_symbols(alphabet)



import math
import time
from queue import Queue

from trainify.validator.pyModelChecking import *
from trainify.validator.pyModelChecking.CTL import Parser, modelcheck


class FMValidator:
    def __init__(self, configurator, net, logger):

        self.initial_states = configurator.initial_state
        self.initial_states_region = configurator.initial_state_region
        self.proposition_list = configurator.proposition_list
        self.limited_count = configurator.limited_count
        self.limited_depth = configurator.limited_depth
        self.atomic_propositions = configurator.atomic_propositions
        self.formula = configurator.formula
        # 具体状态转为抽象状态
        self.get_abstract_state = configurator.get_abstract_state
        self.get_abstract_state_label = configurator.get_abstract_state_label
        self.get_abstract_state_hash = configurator.get_abstract_state_hash
        self.get_next_states = configurator.get_next_states
        self.is_done = configurator.is_done
        self.get_low_dim_state = configurator.get_low_dim_state
        # 抽象初始状态列表
        self.abstract_initial_states = []

        self.network = net
        self.divide_tool = configurator.divide_tool
        # 临时状态

        # Angle at which to fail the episode
        self.theta_threshold_radians = 12 * 2 * math.pi / 360
        self.x_threshold = 2.4

        # debug
        self.visited_list = []
        self.logger = logger

    def get_initial_state(self):
        self.abstract_initial_states = []
        if self.initial_states_region is not None:
            initial_states = self.divide_tool.intersection(self.initial_states_region)
            for s in initial_states:
                self.abstract_initial_states.append(s)
        else:
            self.abstract_initial_states.append(self.get_abstract_state(self.initial_states))
        self.logger.info('number of initial states:' + str(len(self.abstract_initial_states)))

    def create_kripke_ctl(self):
        label_dict = {}
        all_des_set = set()
        abstract_state_count = 0
        sta_cnt = 0
        self.get_initial_state()
        bfs = Queue()
        # 初始状态添加标签
        for abstract_state in self.abstract_initial_states:
            low_dim = self.get_low_dim_state(abstract_state)
            all_des_set.add(abstract_state)
            label_dict[low_dim] = self.get_abstract_state_label(low_dim, 0)
            bfs.put(abstract_state)
            abstract_state_count += 1
            sta_cnt += 1
            # self.logger.info("初始抽象状态"+ str(abstract_state))

        edges = []
        edge_set = set()
        tmp = Queue()
        old_abs_cnt = 1

        for i in range(self.limited_depth):
            bad_label_cnt = 0
            bad_state = False
            t0 = time.time()
            tt = 0
            if bfs.empty():
                break
            while not bfs.empty():
                # 取出同一层的所有节点
                node = bfs.get()
                tmp.put(node)
            # 计算该层所有节点的后继节点
            # mem_used = float(psutil.virtual_memory().used) / 1073741824
            while not tmp.empty():
                current = tmp.get()
                cur_low_dim = self.get_low_dim_state(current)
                t1 = time.time()
                des = self.get_next_states(current)
                t2 = time.time()
                tt = tt + t2 - t1
                for j in range(len(des)):
                    des_low_dim = self.get_low_dim_state(des[j])
                    if not (des[j] in all_des_set):
                        all_des_set.add(des[j])
                        bfs.put(des[j])
                        sta_cnt += 1
                    index = label_dict.get(des_low_dim, -1)
                    if not ((cur_low_dim, des_low_dim) in edge_set):
                        edges.append((cur_low_dim, des_low_dim))
                        edge_set.add((cur_low_dim, des_low_dim))
                    if index == -1:
                        s_label = self.get_abstract_state_label(des_low_dim, bad_label_cnt)
                        label_dict[des_low_dim] = s_label
                        # if len(s_label) == 0:
                        #     if bad_label_cnt <= 1:
                        #         self.logger.info('current state-----------'+ str(current))
                        #     bad_state = True
                        #     bad_label_cnt += 1
                        abstract_state_count += 1
            t3 = time.time()
            self.logger.info(str(abstract_state_count) + str(sta_cnt) + 'depth' + str(i) + str(t3 - t0) + ' ' + str(
                tt) + 'increase' +
                             str(abstract_state_count - old_abs_cnt))
            # if bad_state:
            #     return None
            old_abs_cnt = abstract_state_count
            # self.logger.info(str(abstract_state_count)+ 'depth'+str( i)+str( t3 - t0)+ ' '+ str(tt)+ '--'+ str(mlp)+ '---'+str( rmp))

        self.logger.info("final_state_count: " + str(abstract_state_count))
        for i in range(len(self.visited_list) - 1):
            if not (self.visited_list[i], self.visited_list[i + 1]) in edges:
                self.logger.error('error')
        k = Kripke(R=edges, L=label_dict)
        self.logger.info(str(k.labels()))

        return k

    def ctl_model_check(self, k):
        parser = Parser()
        f = parser(self.formula)
        qualified_states = modelcheck(k, f)
        s_n = len(self.abstract_initial_states)
        self.logger.info(str(s_n) + self.formula)
        # 验证qualified_states是否包含所有初始状态
        flag = True
        sat = 0
        unsat = 0
        for i in range(len(self.abstract_initial_states)):
            # 如果存在初始状态不满足该性质，则可以判定系统不满足该性质
            low_dim_state = self.get_low_dim_state(self.abstract_initial_states[i])
            if low_dim_state not in qualified_states:
                flag = False
                unsat += 1
            else:
                sat += 1

        if flag:
            self.logger.info("CTL formula satisfied:" + str(self.formula) + str(sat) + str(unsat))
            return True, qualified_states

        else:
            self.logger.info("CTL formula not satisfied:" + str(self.formula) + str(sat) + str(unsat))
            if unsat < s_n:
                return True, qualified_states
            else:
                return False, qualified_states



import copy
import multiprocessing
import time

import torch
from scipy.optimize import minimize, Bounds
from trainify.abstract.divide_tool import str_to_list, max_min_clip, combine_bound_list, near_bound, \
    initiate_divide_tool
from trainify.BBReach.draw import *
import re

from trainify.agent import Actor as ddpgActor


def do_BBReach(actor_network, recorder, verify_config=None, env_config=None):
    if env_config is None:
        env_config = {
            "dim": 3,
            "states_name": ['x1', 'x2', 'x3'],
            "state_space": [[-2.5, -2.5, -2.5], [2.5, 2.5, 2.5]],
            "abs_initial_intervals": [0.5, 0.5, 0.5],
            "state_key_dim": [0, 1],
            "dynamics": ['x[0] + (-x[0] + x[1] - x[2]) * 0.02', 'x[1] + (-x[0] * (x[2] + 1) - x[1]) * 0.02',
                         'x[2] + (-x[0] + action) * 0.02'],
        }
    if verify_config is None:
        verify_config = {
            "distance_threshold": [0.001, 0.0001, 0.0001],
            "initial_set": [0.25, 0.08, 0.25, 0.27, 0.1, 0.27],
            "max_step": 35,
            "initial_set_partition": [0.01, 0.01, 0.02]
        }
    divide_tool = initiate_divide_tool(env_config['state_space'], env_config['abs_initial_intervals'])
    reach_env = ReachEnv("b4_env", divide_tool=divide_tool, network=actor_network)
    reach_env.state_space = env_config['state_space']
    reach_env.xnames = env_config['states_name']
    reach_env.cd = env_config['dynamics']
    # TODO
    reach_env.standard = verify_config['distance_threshold']
    # r = [0.25, 0.08, 0.25, 0.27, 0.1, 0.27]
    res_list = calculate_reachable_sets(reach_env, verify_config['initial_set'], verify_config['max_step'],
                                        logger=recorder.logger)
    draw_box(res_list, recorder)
    # parallel_list = parallel_cal(reach_env, verify_config['initial_set'], verify_config['initial_set_partition'],
    #                              verify_config['max_step'],
    #                              process_num=1)
    # draw_box(parallel_list, recorder, parallel=True)
    recorder.logger.info('finished')

    return 0


def create_function(str, max):
    labels = [
        ['sin', 'math.sin'],
        ['cos', 'math.cos'],
        ['tan', 'math.tan'],
        ['tanh', 'math.tanh']
    ]
    for a in labels:
        reg = re.compile(re.escape(a[0]), re.IGNORECASE)
        str = reg.sub(a[1], str)

    def cal(x):
        if max:
            return -eval(str)
        else:
            return eval(str)

    return cal


def calculate_reachable_sets(env, initial_bound, time_step, identifier=None, logger=None):
    if isinstance(initial_bound, str):
        initial_bound = str_to_list(initial_bound)
    actual_dim = len(env.cd)
    assert len(initial_bound) == 2 * actual_dim
    bound_list = [initial_bound]
    res_list = [[initial_bound[0], initial_bound[1], initial_bound[0 + actual_dim], initial_bound[1 + actual_dim]]]
    env.timer_reset()
    st = time.time()
    t = 0
    while True:
        t0 = time.time()
        bound_list = env.get_next_bound_list(bound_list)
        min_x1 = math.inf
        max_x1 = -math.inf
        min_x2 = math.inf
        max_x2 = -math.inf
        for bound in bound_list:
            min_x1 = min(bound[0], min_x1)
            max_x1 = max(bound[0 + actual_dim], max_x1)
            min_x2 = min(bound[1], min_x2)
            max_x2 = max(bound[1 + actual_dim], max_x2)
        res_list.append([min_x1, min_x2, max_x1, max_x2])

        t1 = time.time()
        if identifier is None:
            logger.info(
                str(t) + '：' + str(len(bound_list)) + str(min_x1) + str(max_x1) + str(min_x2) + str(max_x2) + str(
                    t1 - t0))
        t += 1
        if t == time_step:
            if identifier is not None:
                logger.info(
                    str(identifier) + '：' + 'min_x1,' + str(min_x1) + 'max_x1,' + str(max_x1) + 'min_x2,' + str(
                        min_x2) + 'max_x2,' + str(max_x2))
            break
    et = time.time()
    logger.info('Overall Time' + str(et - st))
    logger.info('seg' + str(env.time_seg) + 'over-app' + str(env.time_op) + 'agg' + str(env.time_agg))
    return np.array(res_list)


def err_call_back(err):
    print('error---', err)


def parallel_cal(env, initial_bound, set_partition_gran, time_step, process_num=4):
    st = time.time()
    half_dim = len(env.cd)
    assert len(initial_bound) == 2 * len(env.cd)
    assert len(set_partition_gran) == len(env.cd)
    sr = [initial_bound[0:half_dim], initial_bound[half_dim:]]
    sr_dt = initiate_divide_tool(sr, set_partition_gran)
    bounds = sr_dt.intersection(initial_bound)
    print('Number of sub-tasks', len(bounds))
    results = []
    pool = multiprocessing.Pool(processes=process_num)
    cnt = 1
    for bound in bounds:
        results.append(
            pool.apply_async(calculate_reachable_sets, args=(env, bound, time_step, cnt), error_callback=err_call_back))
        cnt += 1
    pool.close()
    pool.join()
    parallel_res_list = []
    for res in results:
        p_list = res.get()
        parallel_res_list.append(p_list)
        # print(res)
    dim_list = list(range(4))
    parallel_res_list = np.array(parallel_res_list)
    max_res = parallel_res_list.max(axis=0)
    max_res = np.delete(max_res, dim_list[0:2], axis=1)
    min_res = parallel_res_list.min(axis=0)
    min_res = np.delete(min_res, dim_list[2:4], axis=1)
    r = np.append(min_res, max_res, axis=1)
    et = time.time()
    print('Time of Parallel Calculation:', et - st)
    return np.array(r)


class ReachEnv():
    def __init__(self, name, env_config={}, divide_tool=None, network=None):
        self.name = name
        self.xnames = ['x1', 'x2']
        # self.d = ["x[1] + x[2]*0.1", "x[2]+(x[0]*x[2]*x[2]-x[1])*0.1"]
        self.cd = ["x[0] + x[1]*0.1", "x[1]+(action*x[1]*x[1]-x[0])*0.1"]
        self.state_space = [[-2, -2], [2, 2]]
        self.verify_func = []
        # self.create_verify_func()
        self.divide_tool = divide_tool
        self.network = network
        self.standard = [0.0001, 0.0001]
        self.tau = 0.1
        self.state_dim = len(self.cd)

        self.time_seg = 0
        self.time_agg = 0
        self.time_op = 0

    def timer_reset(self):
        self.time_seg = 0
        self.time_agg = 0
        self.time_op = 0

    # def create_verify_func(self):
    #     labels = [
    #         ['sin', 'math.sin'],
    #         ['cos', 'math.cos'],
    #         ['tan', 'math.tan'],
    #         ['tanh', 'math.tanh']
    #     ]
    #
    #     def from_function(str, max):
    #         def cal(x):
    #             if max:
    #                 return -eval(str)
    #             else:
    #                 return eval(str)
    #
    #         return cal
    #
    #     for i, str in enumerate(self.d):
    #         for a in labels:
    #             reg = re.compile(re.escape(a[0]), re.IGNORECASE)
    #             self.d[i] = reg.sub(a[1], self.d[i])
    #
    #         setattr(self, self.xnames[i] + '_maximum',
    #                 from_function(self.d[i], True))
    #         setattr(self, self.xnames[i] + '_minimum',
    #                 from_function(self.d[i], False))
    #
    #         self.verify_func.append(from_function(self.d[i], False))
    #         self.verify_func.append(from_function(self.d[i], True))

    def get_next_bound_list(self, bound_list):
        res_list = []
        cnt = 0
        for bound in bound_list:
            next_bound_list, counter = self.get_next_states(bound)
            cnt += counter
            for next_bound in next_bound_list:
                res_list.append(next_bound)
        t1 = time.time()
        u = combine_bound_list(res_list, self.standard)
        t2 = time.time()
        self.time_agg = self.time_agg + t2 - t1
        # print(cnt)
        return u

    # Method must be implemented by users
    def get_next_states(self, current):
        if isinstance(current, str):
            current = str_to_list(current)
        current = copy.deepcopy(current)
        t0 = time.time()
        target_list = self.divide_tool.intersection(current)
        # print(len(target_list))
        tl = []
        for ele in target_list:
            s = str_to_list(ele)
            s = max_min_clip(current, s)
            tl.append(s)
        t1 = time.time()
        self.time_seg = self.time_seg + t1 - t0
        # over-approximation
        b_list = []
        for t in tl:
            b = self.gn(t)
            b_list.append(b)
        t2 = time.time()
        self.time_op = self.time_op + t2 - t1
        # aggregation
        res = combine_bound_list(b_list, self.standard)
        t3 = time.time()
        self.time_agg = self.time_agg + t3 - t2
        return res, len(b_list)

    def gn(self, current):
        cur_list = self.divide_tool.intersection(current)
        original = None
        dim = len(current)
        half_dim = int(dim / 2)
        for cu in cur_list:
            cu = str_to_list(cu)
            flag = True
            for i in range(half_dim):
                if cu[i] > current[i] or cu[i + half_dim] < current[i + half_dim]:
                    flag = False
                    break
            if flag:
                original = cu
        assert (original is not None)

        s0 = torch.tensor(original, dtype=torch.float).unsqueeze(0)
        action = self.network(s0).squeeze(0).detach().numpy()
        offset = 0
        scala = 1
        self.action = (action[0] - offset) * scala
        next_bounds = self.execute_action(current)

        # action 通过约束传递求解很慢
        # mid_point = [self.action]
        # lb = []
        # ub = []
        # for i in range(half_dim):
        #     mid_point.append((current[i] + current[i + half_dim]) / 2)
        #     lb.append(current[i])
        #     ub.append(current[i + half_dim])
        # bound = Bounds(lb, ub)
        # cons_list = []
        # action_cons = {'type': 'eq', 'fun': lambda x: x[0] - self.action}
        # cons_list.append(action_cons)
        #
        # # 循环内的lambda表达式，坑。。在lambda函数中添加默认参数循环索引
        # for j in range(half_dim):
        #     state_lb_cons = {'type': 'ineq', 'fun': lambda x, j=j: x[j + 1] - lb[j]}
        #     cons_list.append(state_lb_cons)
        #     state_ub_cons = {'type': 'ineq', 'fun': lambda x, j=j: ub[j] - x[j + 1]}
        #     cons_list.append(state_ub_cons)
        #
        # next_bounds = [0 for i in range(dim)]

        # for i in range(half_dim):
        #     l = minimize(self.verify_func[2 * i], x0=mid_point, method='SLSQP', constraints=cons_list)
        #     l = self.verify_func[2 * i](l.x)
        #     r = minimize(self.verify_func[2 * i + 1], x0=mid_point, method='SLSQP', constraints=cons_list)
        #     r = -self.verify_func[2 * i + 1](r.x)
        #     # TODO state_bound
        #     l = np.clip(l, -2, 2)
        #     r = np.clip(r, -2, 2)
        #     next_bounds[i] = l
        #     next_bounds[i + half_dim] = r

        # 使用区间运算
        # interval_bound = []
        # for j in range(half_dim):
        #     ele = interval[lb[j], ub[j]]
        #     interval_bound.append(ele)
        #
        # next_bounds = [0 for i in range(dim)]
        # for i in range(half_dim):
        #     fun_str = self.cd[i]
        #     fun_str = fun_str.replace('action', str(self.action))
        #     lb_func = create_function(fun_str, False)
        #     next_interval_bound = lb_func(interval_bound)
        #     l = next_interval_bound[0][0]
        #     r = next_interval_bound[0][1]
        #     # TODO state_bound
        #     l = np.clip(l, -2, 2)
        #     r = np.clip(r, -2, 2)
        #     next_bounds[i] = l
        #     next_bounds[i + half_dim] = r

        return next_bounds

    def execute_action(self, current):
        dim = len(current)
        half_dim = int(dim / 2)
        # 将action直接通过字符串替换成常数，避免添加action的约束
        mid_point = []
        lb = []
        ub = []
        for i in range(half_dim):
            mid_point.append((current[i] + current[i + half_dim]) / 2)
            lb.append(current[i])
            ub.append(current[i + half_dim])
        bound = Bounds(lb, ub)
        cons_list = []

        # 循环内的lambda表达式，坑。。在lambda函数中添加默认参数循环索引
        for j in range(half_dim):
            state_lb_cons = {'type': 'ineq', 'fun': lambda x, j=j: x[j] - lb[j]}
            cons_list.append(state_lb_cons)
            state_ub_cons = {'type': 'ineq', 'fun': lambda x, j=j: ub[j] - x[j]}
            cons_list.append(state_ub_cons)

        next_bounds = [0 for i in range(dim)]
        for i in range(half_dim):
            fun_str = self.cd[i]
            fun_str = fun_str.replace('action', str(self.action))
            lb_func = create_function(fun_str, False)
            ub_func = create_function(fun_str, True)
            l = minimize(lb_func, x0=mid_point, method='COBYLA', constraints=cons_list)
            l = lb_func(l.x)
            r = minimize(ub_func, x0=mid_point, method='COBYLA', constraints=cons_list)
            r = -ub_func(r.x)
            l = np.clip(l, self.state_space[0][i], self.state_space[1][i])
            r = np.clip(r, self.state_space[0][i], self.state_space[1][i])
            next_bounds[i] = l
            next_bounds[i + half_dim] = r
        return next_bounds


if __name__ == "__main__":
    # env = ReachEnv("b1_env")
    # # l = minimize(env.verify_func[0], x0=[-0.02639258, 0.895, 0.535], method='SLSQP', constraints=cons)
    # gefunc = env.x2_minimum([0.96370608, 0.9, 0.54])
    # res = env.verify_func[2]([-0.96370608, 0.9, 0.54])
    # x0 = interval[0.89, 0.9]
    # x1 = interval[0.53, 0.54]
    # x = [x0, x1]
    # test_func = create_function("x[1]+(-1.9954697*x[1]*x[1]-x[0])*0.1", False)
    # tt = test_func(x)
    # self.network.load_state_dict(torch.load(pt_file0))

    # B1
    # pt_file = "b1_abs-actor_[0.05, 0.05]_2_20.pt"
    # network = ddpgActor(4, 20, 1)
    # network.load_state_dict(torch.load(pt_file))
    # state_space = [[-2.5, -2.5], [2.5, 2.5]]
    # initial_intervals = [0.05, 0.05]
    # divide_tool = initiate_divide_tool(state_space, initial_intervals)
    # b1 = ReachEnv("b1_env", divide_tool=divide_tool, network=network)
    # r = [0.8, 0.5, 0.9, 0.6]
    # # r = [0.89, 0.53, 0.9, 0.54]
    # calculate_reachable_sets(b1, r, 60)
    # parallel_cal(b1, r, [0.05, 0.1], 60, process_num=2)

    # B4
    state_space = [[-2.5, -2.5, -2.5], [2.5, 2.5, 2.5]]
    initial_intervals = [0.5, 0.5, 0.5]
    pt_file = "b4_abs-actor_" + str(initial_intervals) + "_2_20.pt"
    network = ddpgActor(6, 20, 1)
    network.load_state_dict(torch.load(pt_file))
    divide_tool = initiate_divide_tool(state_space, initial_intervals)
    b4 = ReachEnv("b4_env", divide_tool=divide_tool, network=network)
    b4.state_space = state_space
    b4.xnames = ['x1', 'x2', 'x3']
    b4.cd = ['x[0] + (-x[0] + x[1] - x[2]) * 0.02', 'x[1] + (-x[0] * (x[2] + 1) - x[1]) * 0.02',
             'x[2] + (-x[0] + action) * 0.02']
    b4.standard = [0.001, 0.0001, 0.0001]
    r = [0.25, 0.08, 0.25, 0.27, 0.1, 0.27]
    res_list = calculate_reachable_sets(b4, r, 35)
    draw_box(res_list, False)
    parallel_list = parallel_cal(b4, r, [0.01, 0.01, 0.02], 35, process_num=2)
    draw_box(parallel_list)
    print('finished')



"""
.. module:: language
   :synopsis: Represents formulas and logics.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

import sys
import inspect


class Formula(object):
    ''' A class to represent formulas.

    Formulas are represented as nodes in labelled trees: leaves are terminal
    symbols (e.g., atomic propositions and Boolean values), while internal
    nodes correspond to operators and quantifiers. The ariety of internal nodes
    depends on the kind of operator or quantifier must be represented.
    For instance, the arity of a node representing the formula
    :math:`not (p \lor True)` is one because the formula has exclusively one
    sub-formula, i.e., :math:`p \lor True`. On the contrary, this last formula
    has two sub-formulas, i.e., :math:`p` and  :math:`True`, thus, the node
    representing it has two sons.
    '''

    __desc__ = 'formula'

    def __init__(self, *phi):
        self.wrap_subformulas(phi, sys.modules[self.__module__].Formula)

    def wrap_subformulas(self, subformulas, FormulaClass):
        ''' Replaces subformulas of the current object

        This method replaces the subformulas of the current object by
        using the :class:`FormulaClass` objects representing the
        elements of :param subformulas:.

        :param self: this object
        :type self: Formula
        :param subformulas: the set of objects to be used as model for the
            replacement
        :type subformulas: Container
        :param FormulaClass: the final class of the new subformulas
        :type FormulaClass: class
        '''

        def err_msg(phi):
            return ('expected a {}, '.format(FormulaClass.__desc__) +
                    'got the {} {}'.format(phi.__desc__, phi))

        Lang = sys.modules[self.__module__]

        self._subformula = []
        self.height = 0

        for phi in subformulas:
            if isinstance(phi, bool):
                self._subformula.append(Lang.Bool(phi))
            else:
                if not isinstance(phi, FormulaClass):
                    if (isinstance(phi, Lang.Formula) or
                            not isinstance(phi, Formula)):

                        raise TypeError(err_msg(phi))

                    phi = phi.cast_to(Lang)

                    if not isinstance(phi, FormulaClass):
                        raise TypeError(err_msg(phi))

                self._subformula.append(phi)
                self.height = max(self.height, phi.height+1)

    def clone(self):
        ''' Clones a formula

        :returns: a clone of the current formula
        :rtype: Formula
        '''
        return self.__class__(*[sf.clone() for sf in self._subformula])

    def __and__(self, obj):
        Lang = sys.modules[self.__module__]

        return Lang.And(self, obj)

    def __rand__(self, obj):
        Lang = sys.modules[self.__module__]

        return Lang.And(obj, self)

    def __or__(self, obj):
        Lang = sys.modules[self.__module__]

        return Lang.Or(self, obj)

    def __ror__(self, obj):
        Lang = sys.modules[self.__module__]

        return Lang.Or(obj, self)

    def __invert__(self):
        Lang = sys.modules[self.__module__]

        return Lang.Not(self)

    def __hash__(self):
        return str(self).__hash__()

    def __eq__(self, other):
        return str(self) == str(other)

    def __cmp__(self, other):
        self_str = str(self)
        other_str = str(other)
        if (self_str < other_str):
            return -1

        if (self_str > other_str):
            return 1

        return 0

    def subformula(self, i):
        ''' Returns the :math:`i`-th subformula.

        :param i: the index of the subformula to be returned
        :type i: Integer
        :returns: returns the :math:`i`-th subformula of the current formula
        :rtype: Formula
        '''
        return self._subformula[i]

    def subformulas(self):
        ''' Returns the list of all the subformulas.

        :returns: returns the list of the subformulas of the current formula
        :rtype: list
        '''
        return self._subformula

    def __repr__(self):
        return str(self)


class Bool(Formula):
    '''
    The class of Boolean values.

    '''

    symbols = {True: 'true',
               False: 'false'}

    def __init__(self, value):
        ''' Initialize a Boolean value.

        This method builds either a True or a False value.

        :param value: the boolean value.
        :type value: bool

        '''
        if not isinstance(value, bool):
            raise TypeError('\'{}\' must be boolean value'.format(value))
        self._value = value
        self.height = 0

    def clone(self):
        ''' Clones an atomic proposition

        :returns: a clone of the current Bool value
        :rtype: Bool
        '''
        return self.__class__(bool(self._value))

    def subformula(self, i):
        ''' Returns the :math:`i`-th subformula.

        :param i: the index of the subformula to be returned
        :type i: Integer
        :raise TypeError: Boolean values have not subformulas
        '''
        raise TypeError('Bools have not subformulas.')

    def subformulas(self):
        ''' Returns the list of all the subformulas.

        :returns: returns the empty list of the subformulas of the current
            formula
        :rtype: list
        '''
        return []

    def __hash__(self):
        return str(self).__hash__()

    def __eq__(self, other):

        if isinstance(other, bool):
            return self._value == other

        if isinstance(other, Bool):
            return self._value == other._value

        return False

    def __str__(self):
        ''' Return a string depicting the Boolean value.

        :returns: a string depicting the current Boolean value.
        :rtype: str
        '''
        return Bool.symbols[self._value]


class AlphabeticSymbol(object):
    pass


class LogicOperator(Formula):
    '''
    A class to represent logic operator such as :math:`\land` or :math:`\lor`.

    '''
    def __str__(self):
        if len(self._subformula) == 1:
            return '{} {}'.format(self.__class__.symbols[0],
                                  self._subformula[0])
        else:
            sep = ' {} '.format(self.__class__.symbols[0])
            return '({})'.format(sep.join([str(f) for f in self._subformula]))


class Not(LogicOperator, AlphabeticSymbol):
    '''
    Represents logic negation.

    '''

    symbols = ['not', '~']

    def __init__(self, phi):
        super(Not, self).__init__(phi)


class Or(LogicOperator, AlphabeticSymbol):
    '''
    Represents logic non-exclusive disjunction.

    '''
    symbols = ['or', '|']


class And(LogicOperator, AlphabeticSymbol):
    '''
    Represents logic conjunction.

    '''
    symbols = ['and', '&']


class Imply(LogicOperator, AlphabeticSymbol):
    '''
    Represents logic implication.

    '''
    symbols = ['-->']

    def __init__(self, phi, psi):
        super(Imply, self).__init__(phi, psi)


def get_alphabet(name):
    alphabet = {}
    for name, obj in inspect.getmembers(sys.modules[name], inspect.isclass):
        if obj != AlphabeticSymbol and issubclass(obj, AlphabeticSymbol):
            alphabet[name] = obj

    return alphabet


def get_symbols(alphabet):
    symbols = list()

    for name, c in alphabet.items():
        if name == 'Bool':
            symbols.extend(Bool.symbols.values())
        else:
            if name != 'AtomicProposition':
                symbols.extend(c.symbols)

    return symbols


def LNot(formula):
    ''' Logical Not - Returns an equivalent formula that does not begin
    with two negations.

    This method negates the parameter and removes all the pairs of negations at
    the begin of the obtained formula.

    :param formula: a formula
    :type formula: Formula
    :returns: a formula equivalent to the parameter that does not begin with
        two negations
    :rtype: Formula
    '''

    if isinstance(formula, Not):
        if isinstance(formula.subformula(0), Not):
            return LNot(formula.subformula(0).subformula(0))
        return formula.subformula(0)

    return sys.modules[formula.__module__].Not(formula)


alphabet = get_alphabet(__name__)
symbols = get_symbols(alphabet)



from torch.utils.tensorboard import SummaryWriter
import subprocess
import os
import threading

def writeTensorBoard(directory, title, data):
    """
    :param directory: str,tensorboard文件的存储目录
    :param title: str,数据标题
    :param data: 一维列表,要写到tensorboard中的数据
    :return:
    """
    writer = SummaryWriter(log_dir= directory)
    for i in range(len(data)):
        writer.add_scalar(tag= title, scalar_value= data[i], global_step=i)
    writer.close()

def _openTensorBoard(path, directory):
    """
    :param path: 当前工程路径
    :param directory: tensorboard文件的存储目录名称
    :return:
    """
    if os.name == "nt":
        # windows
        cmds = [path.split("\\")[0], "cd " + path, "tensorboard --logdir=" + directory]
        p = subprocess.Popen('cmd.exe', stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        for cmd in cmds:
            p.stdin.write((cmd + "\n").encode(encoding='UTF-8', errors='strict'))
        p.stdin.close()
        print(p.stdout.read())
    elif os.name == "posix":
        # linux
        cmds = ["cd " + path, "tensorboard --logdir=" + directory]
        p = subprocess.Popen('shell', stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        for cmd in cmds:
            p.stdin.write((cmd + "\n").encode(encoding='UTF-8', errors='strict'))
        p.stdin.close()
        print(p.stdout.read())
    else:
        # mac
        pass

def openTensorBoard(path, directory):
    t1 = threading.Thread(target=_openTensorBoard, args=(path,directory,))
    t1.start()




import os
import sys

from trainify.abstract.pendulum.pendulum_abs import *
from trainify.validator.cegar import cegar
from trainify.env.verify.pendulum_env import PendulumEnv

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(BASE_DIR)

# def cegar(file_name, agent, state_space, initial_intervals, train_model, verify_env, max_iteration):
file_name = 'pendulum_Relu_3_128_p3'
initial_intervals = [0.16, 0.16, 0.01]

divide_tool = initiate_divide_tool_rtree(state_space, initial_intervals, [0, 1], file_name)
agent = Agent(divide_tool)
pd = PendulumEnv(divide_tool, agent.actor)

cegar(file_name, agent, divide_tool, train_model, pd, 5)



"""
.. module:: PL
   :synopsis: Represents the propositional logics.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from .language import *
from ..language import LNot

from .parser import Parser



import os
import time
import sys
import torch
from tensorboardX import SummaryWriter
import subprocess
from trainify.data.logger import Logger
from trainify.utils import send_to_backend

ROOT_PROJECT_PATH = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
ROOT_DATA_PATH = os.path.join(ROOT_PROJECT_PATH, "data")


class Recorder:
    def __init__(self,
                 experiment_name,
                 result_dir_name,
                 result_path,
                 backend_channel
                 ):
        self.experiment_name = experiment_name
        self.result_dir_name = result_dir_name
        self.backend_channel = backend_channel if backend_channel is not None else False
        # print(self.data_path)
        self.data_dir = result_path + '/' + self.result_dir_name
        self.data_dir_log = self.data_dir + '/log'
        self.data_dir_file = self.data_dir + '/files'
        self.data_dir_tensorboard = self.data_dir + '/tensorboard'

        self.Logger = Logger(log_path=self.data_dir_log, channel=self.backend_channel)
        self.logger = self.Logger.create_logger(logger_name='log_' + self.experiment_name)

        self._create_dir(self.data_dir)
        self._create_dir(self.data_dir_log)
        self._create_dir(self.data_dir_file)
        self._create_dir(self.data_dir_tensorboard)

        self.chart = {}
        self.file_list = []
        self.reward = {}

        self.logger.info('Recorder init success')

    def _create_dir(self, path):
        if not os.path.exists(path):
            os.makedirs(path)
        self.logger.info('Recorder 数据存储文件夹创建完成：' + path)
        return None

    def create_chart(self, name, desc):
        self.chart[name] = {
            'x': [],
            'y': [],
            'desc': desc
        }

    def add_chart(self, name, data):
        self.chart[name]['x'].append(data['x'])
        self.chart[name]['y'].append(data['y'])

        if self.backend_channel:
            data[name] = name
            data['desc'] = self.chart[name]['desc']
            content = {
                "type": "chart",
                "data": data
            }
            send_to_backend(self.backend_channel, content)

    def add_file(self, type, data):
        if type.startswith("file/"):
            content = {
                "type": type,
                "data": data
            }
            find_it = None
            for i, v in enumerate(self.file_list):
                if v.get('type') == type and v.get('data').get('file_path') == data.get('file_path'):
                    find_it = i
                    self.file_list[i] = content
                    return
            if find_it is None:
                self.file_list.append(content)
                if self.backend_channel:
                    send_to_backend(self.backend_channel, content)

    def create_experiment(self, title):
        self.temp_name = title
        self.reward.update({title: {'name': title, 'reward_list': []}})

    def get_data_path(self):
        return self.data_dir

    def add_reward(self, reward, title=''):
        if title == '':
            title = self.temp_name
        self.reward[title]['reward_list'].append(reward)

    def get_reward_list(self, title=''):
        if title == '':
            title = self.temp_name
        return self.reward[title]['reward_list']

    def get_reward(self, title=''):
        if title == '':
            title = self.temp_name
        return self.reward[title]

    def writeAll2TensorBoard(self, title=''):
        """
        :param title: str,数据标题
        :return:
        """
        if title == '':
            title = self.temp_name

        writer = SummaryWriter(log_dir=self.data_dir_tensorboard)
        for title in self.reward.keys():
            for i in range(len(self.reward[title]['reward_list'])):
                writer.add_scalar(tag=title, scalar_value=self.reward[title]['reward_list'][i], global_step=i)
        writer.close()

    def openTensorBoard(self):
        """
        :param path: 当前工程路径
        :param directory: tensorboard文件的存储目录名称
        :return:
        """
        cmds = [ROOT_PROJECT_PATH.split("\\")[0], "cd " + ROOT_PROJECT_PATH, "tensorboard --logdir=data"]
        p = subprocess.Popen('cmd.exe', stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        for cmd in cmds:
            p.stdin.write((cmd + "\n").encode(encoding='UTF-8', errors='strict'))
        p.stdin.close()
        print(p.stdout.read())

    def create_save_model(self, agent, agent_config):
        def save_model():
            for name in agent_config['models_need_save']:
                if hasattr(agent, name):
                    torch.save(agent.__dict__[name].state_dict(), self.data_dir_file + '/' + name + '.pt')
                    self.add_file("file/pt", {
                        "file_path": self.data_dir_file + '/' + name + '.pt',
                        "info": "训练时保存的模型"
                    })
                    self.logger.info('模型 ' + name + ' 保存完毕')

        return save_model

    def create_load_model(self, agent, agent_config):
        def load_model():
            for name in agent_config['models_need_save']:
                if hasattr(agent, name):
                    agent.__dict__[name].load_state_dict(torch.load(self.data_dir_file + '/' + name + '.pt'))
                    self.logger.info('Trainify 模型 ' + name + ' 加载完毕')

        return load_model


if __name__ == '__main__':
    a = Recorder('ttt', 'ttt_results')
    a.create_experiment("test1")
    for i in range(100):
        a.add_reward("test1", i * 6)
    a.create_experiment("test2")
    for i in range(100):
        a.add_reward("test2", i * 10)
    a.writeAll2TensorBoard()
    a.openTensorBoard()



from ..parser import Parser as BaseParser
from ..parser import AST_to_Logics


class AST_to_PropositionalLogics(AST_to_Logics):
    ''' A class to transform AST into a propositional formula.

    This class implements a Lark's transformer to turn AST reprensenting
    propositional logics formulas into the corresponding propositional formula.
    '''

    def __init__(self, lang):
        super(AST_to_PropositionalLogics, self).__init__(lang)

    def a_prop(self, subformula):
        return subformula[0]

    def string(self, s):
        return self.__lang__.AtomicProposition(str(s[0]))

    def e_string(self, s):
        return self.__lang__.AtomicProposition(str(s[0])[1:-1])

    def or_formula(self, subformulas):
        return self.__lang__.Or(*subformulas)

    def and_formula(self, subformulas):
        return self.__lang__.And(*subformulas)

    def imply_formula(self, subformulas):
        return self.__lang__.Imply(*subformulas)

    def not_formula(self, subformulas):
        return self.__lang__.Not(*subformulas)

    def true(self, subformula):
        return self.__lang__.Bool(True)

    def false(self, subformula):
        return self.__lang__.Bool(False)

    def b_formula(self, subformula):
        return subformula[0]

    def s_formula(self, subformula):
        return subformula[0]

    def u_formula(self, subformula):
        return subformula[0]

    def formula(self, subformula):
        return subformula[0]


class Parser(BaseParser):
    ''' A class to parse a propositional formula.

    Every object of this class instanciates a Lark's parser to parse a
    string reprensenting propositional formulas into the corresponding
    pyModelChecking's formulas.
    '''

    grammar = r"""
        s_formula: "{}"     -> true
                 | "{}"    -> false
                 | a_prop
                 | "(" s_formula ")"

        u_formula: {} u_formula  -> not_formula
                 | "(" b_formula ")"
                 | s_formula

        b_formula: u_formula
                 | u_formula ( {} u_formula )+ -> or_formula
                 | u_formula ( {} u_formula )+ -> and_formula
                 | u_formula {} u_formula -> imply_formula

        a_prop: /[a-zA-Z_][a-zA-Z_0-9]*/ -> string
              | ESCAPED_STRING           -> e_string

        formula: b_formula

        %import common.ESCAPED_STRING
        %import common.WS
        %ignore WS
        """

    def __init__(self, language=None):
        if language is None:
            import pyModelChecking.PL as PL

            language = PL

        AST_to_PL = AST_to_PropositionalLogics

        super(Parser, self).__init__(grammar=Parser.grammar,
                                     language=language,
                                     AST_Transformer=AST_to_PL)


def init_submodule():
    from .language import symbols, alphabet

    def format_symbols_for(operator):
        return '({})'.format('|'.join(['\"{}\"'.format(s)
                                       for s in alphabet[operator].symbols]))

    Parser.grammar = Parser.grammar.format(alphabet['Bool'].symbols[True],
                                           alphabet['Bool'].symbols[False],
                                           format_symbols_for('Not'),
                                           format_symbols_for('Or'),
                                           format_symbols_for('And'),
                                           format_symbols_for('Imply'))


init_submodule()



from pyModelChecking.CTL import *
import unittest

from .test_CTLS_language import formula_into_str


class TestFormulas(unittest.TestCase):
    def setUp(self):
        self.formulas = [True, 'p', AtomicProposition('q'), AF(EG('q')),
                         AG('p'), EF('p'), AX('p'), AU('q', 'p'), ER('q', 'p'),
                         Imply(AF(EG('q')), AG('p')),
                         AG(Imply(And(Not('Close'), 'Start'),
                                  Or(AG(Not('Heat')), EF(Not('Error')))))]

    def test_simplified_syntax(self):
        formulas = [(And(True, 'p'), Bool(True) & 'p'),
                    (Not(And(True, 'p')), ~(Bool(True) & 'p')),
                    (And(Not(True), 'p'), ~Bool(True) & 'p'),
                    (Or(True, Not('p')), Bool(True) | Not('p')),
                    (Or(True, Not('p')), True | Not('p')),
                    (Or(AG(True), Not(EX('p'))), AG(True) | ~EX('p'))]

        for old_syntax, new_syntax in formulas:
            self.assertEqual(old_syntax, new_syntax)

    def test_atomic_proposition(self):
        s = 'p'
        a = AtomicProposition(s)
        self.assertEqual(s, a)

        self.assertEqual(s, '{}'.format(a))

        with self.assertRaises(TypeError):
            AtomicProposition(1)

    def test_boolean(self):
        for b in [False, True]:
            a = Bool(b)
            self.assertEqual(b, a)
            self.assertEqual(a, b)

        with self.assertRaises(TypeError):
            Bool('a')

    def generic_test_unaryop(self, op, op_str, equivalent_restricted_op=None):
        i = 0
        for phi in self.formulas:
            s = '{} {}'.format(op_str, formula_into_str(phi))
            self.assertEqual(s, '{}'.format(op(phi)))
            if (equivalent_restricted_op is not None and
                    isinstance(phi, Formula)):
                self.assertEqual(op(phi).get_equivalent_restricted_formula(),
                                 equivalent_restricted_op(phi))

    def test_not(self):
        formula = (lambda phi: LNot(phi.get_equivalent_restricted_formula()))
        self.generic_test_unaryop(Not, 'not', formula)

    def test_AF(self):
        formula = (lambda phi:
                   Not(EG(LNot(phi.get_equivalent_restricted_formula()))))
        self.generic_test_unaryop(AF, 'AF', formula)

    def test_EF(self):
        formula = (lambda phi:
                   EU(True, phi.get_equivalent_restricted_formula()))
        self.generic_test_unaryop(EF, 'EF', formula)

    def test_AG(self):
        formula = (lambda phi:
                   Not(EU(True,
                          LNot(phi.get_equivalent_restricted_formula()))))
        self.generic_test_unaryop(AG, 'AG', formula)

    def test_EG(self):
        formula = (lambda phi: EG(phi.get_equivalent_restricted_formula()))
        self.generic_test_unaryop(EG, 'EG', formula)

    def generic_test_binaryop(self, op, middle_op_str,
                              equivalent_restricted_op=None,
                              prefix_op_str=''):
        i = 0
        for phi in self.formulas:
            for psi in self.formulas:
                s = '{}({} {} {})'.format(prefix_op_str, formula_into_str(phi),
                                          middle_op_str, formula_into_str(psi))
                self.assertEqual(s, '{}'.format(op(phi, psi)))
                if (equivalent_restricted_op is not None and
                        isinstance(psi, Formula) and isinstance(phi, Formula)):

                    formula = op(phi, psi).get_equivalent_restricted_formula()
                    self.assertEqual(formula,
                                     equivalent_restricted_op(phi, psi))

    def test_AU(self):
        formula = (lambda phi, psi:
                   Not(Or(EU(LNot(psi.get_equivalent_restricted_formula()),
                             Not(Or(phi.get_equivalent_restricted_formula(),
                                    psi.get_equivalent_restricted_formula()))),
                          EG(LNot(psi.get_equivalent_restricted_formula())))))
        self.generic_test_binaryop(AU, 'U', formula, prefix_op_str='A')

    def test_EU(self):
        formula = (lambda phi, psi:
                   EU(phi.get_equivalent_restricted_formula(),
                      psi.get_equivalent_restricted_formula()))

        self.generic_test_binaryop(EU, 'U', formula, prefix_op_str='E')

    def test_AR(self):
        formula = (lambda phi, psi:
                   Not(EU(LNot(phi.get_equivalent_restricted_formula()),
                          LNot(psi.get_equivalent_restricted_formula()))))
        self.generic_test_binaryop(AR, 'R', formula, prefix_op_str='A')

    def test_ER(self):
        subformula = (lambda phi, psi:
                      Not(Or(LNot(phi.get_equivalent_restricted_formula()),
                             LNot(psi.get_equivalent_restricted_formula()))))
        formula = (lambda phi, psi:
                   Or(EU(psi.get_equivalent_restricted_formula(),
                         subformula(phi, psi)),
                      EG(psi.get_equivalent_restricted_formula())))
        self.generic_test_binaryop(ER, 'R', formula, prefix_op_str='E')

    def test_or(self):
        formula = (lambda phi, psi:
                   Or(phi.get_equivalent_restricted_formula(),
                      psi.get_equivalent_restricted_formula()))
        self.generic_test_binaryop(Or, 'or', formula)

    def test_and(self):
        formula = (lambda phi, psi:
                   Not(Or(LNot(phi.get_equivalent_restricted_formula()),
                          LNot(psi.get_equivalent_restricted_formula()))))
        self.generic_test_binaryop(And, 'and', formula)

    def test_imply(self):
        formula = (lambda phi, psi:
                   Or(LNot(phi.get_equivalent_restricted_formula()),
                      psi.get_equivalent_restricted_formula()))

        self.generic_test_binaryop(Imply, '-->', formula)


if __name__ == '__main__':
    unittest.main()



"""
.. module:: kripke
   :synopsis: A module to represent Kripke structures

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""


from .graph import DiGraph
from .graph import compute_SCCs

from .__init__ import __release__

class Kripke(DiGraph):
    '''
    A class to represent Kripke structures.

    A Kripke structure is a directed graph equipped with a set of initial
    nodes, S0, and a labelling function that maps each node into the set of
    atomic propositions that hold in the node itself. The nodes of Kripke
    structure are called *states*.
    '''

    def __init__(self, S=None, S0=None, R=None, L=None):
        ''' Initialize a new Kripke structure

        :param S: a collection of states
        :type S: a collection
        :param S0: a collection of initial states
        :type S0: a collection
        :param R: a collection of edges
        :type R: a collection
        :param L: a labelling function that maps each state in the set of
                    atomic propositions that hold in the state itself.
        :type L: dict
        '''
        super(Kripke, self).__init__(S, R)

        if S0 is None:
            self.S0 = set()
        else:
            self.S0 = set(self.nodes()) & set(S0)

        pots = set(self.nodes())-set(self.sources())
        #TODO
        # if pots:
        #     raise RuntimeError('R=\'{}\' '.format(R) +
        #                        'is supposed be total (see Kripke ' +
        #                        'definition at ' +
        #                        'https://pymodelchecking.readthedocs.io/en/' +
        #                        'v'+__release__ +
        #                        '/models.html#kripke-structures), ' +
        #                        'but it does not contains as sources ' +
        #                        'the nodes {}'.format(pots))

        if L is None:
            L = dict()

        if not isinstance(L, dict):
            raise RuntimeError('L=\'%s\' must be a dictionary' % (L))

        self._labels = dict()
        for state in self.nodes():
            if state in L:
                try:
                    self._labels[state] = set(L[state])
                except Exception:
                    raise RuntimeError(('L=\'{}\' must be a '.format(L)) +
                                       'dictionary that labels each state ' +
                                       'with a set of atomic propositions')
            else:
                self._labels[state] = set()

    def labelling_function(self):
        ''' Return the labelling function

        :returns: the labelling function
        :rtype: dict
        '''
        return self._labels

    def replace_labelling_function(self, L):
        ''' Replace the labelling function

        :param L: a new labelling function for this Kripke structure
        :type L: dict
        :returns: the former labelling function
        :rtype: dict
        '''
        old_L = self._labels

        self._labels = L

        for s in self.states():
            if s not in self._labels:
                self._labels[s] = set()

        return old_L

    def labels(self, state=None):
        ''' Get the atomic propositions labelling either a state or the whole structure

        :param state: either a state of the Kripke structure or None
        :returns: the atomic propositions that label either a
                  *state*, whenever a parameter *state* is passed, or
                  at least one state of the Kripke structure, otherwise
        :rtype: set
        '''
        if state is not None:
            if state not in self.nodes():
                raise RuntimeError(('state=\'{}\' is '.format(state)) +
                                   'not a state of this Kripke structure')
            return self._labels[state]

        AP = set()
        for ap in self._labels.values():
            AP.update(ap)

        return AP

    def states(self):
        ''' Return the states of a Kripke structure

        :returns: the states of the Kripke structure
        :rtype: set
        '''
        return self.nodes()

    def next(self, src):
        ''' Return the next of a state

        Given a Kripke structure :math:`K=(S,S0,R,L)` and one of its state
        :math:`s`, the *next* of :math:`s` in :math:`K` is the set of all
        those states that are destination of some edges whose source is
        :math:`s` itself i.e., :math:`K.next(s)=\{s' | (s,s') \in R\}`.

        :returns: the set of nodes :math:`\{s' | (s,s') \in R\}`
        :rtype: set
        '''
        try:
            return super(Kripke, self).next(src)
        except Exception:
            raise RuntimeError(('src=\'{}\' is not a state '.format(src)) +
                               'of this Kripke structure')

    def transitions_iter(self):
        ''' Return an interator of the edges of a Kripke structure

        :returns: an interator of the set of edges of the Kripke structure
        :rtype: iterator
        '''
        return self.edges_iter()

    def transitions(self):
        ''' Return the edges of a Kripke structure

        :returns: the set of edges of the Kripke structure
        :rtype: set
        '''
        return self.edges()

    def clone(self):
        ''' Clone a Kripke structure

        :returns: a clone of the current Kripke structure
        :rtype: Kripke
        '''
        L = dict()
        for state, AP in self._labels.items():
            L[state] = set(AP)

        return Kripke(self.states(), self.S0, self.transitions(), L)

    def get_substructure(self, V):
        ''' Return the sub-structure that respects a set of states

        The sub-structure of a Kripke structure :math:`(V',E',L')` that
        respects a set of states :math:`V` is the Kripke structure
        :math:`(V,E,L)` where :math:`E=E'\cap(V \times V)` and
        :math:`L(v)=L'(v)` for all :math:`s \in V`.

        :param V: a set of states
        :type V: set
        :returns: the sub-structure that respects V
        :rtype: Kripke
        '''

        S = V & set(self.states())
        S0 = V & self.S0
        E = [(s, d) for (s, d) in self.transitions_iter() if s in V and d in V]
        L = {s: S for s, S in self._next.items() if s in V}

        return Kripke(S, S0, E, L)

    def get_fair_states(self, F):
        ''' Return a set of states from which leaves a fair path.

        :param F: a container of fairness constraints
        :type F: a container
        :returns: the set of states from which leaves a fair path
        :rtype: set
        '''

        def is_a_fair_SCC(self, scc, F):
            v = next(iter(scc))
            if len(scc) == 1 or v not in self.next(v):
                return False

            for P in F:
                if not set(scc) & P:
                    return False

            return True

        F_set = set()
        for SCC in compute_SCCs(self):
            if is_a_fair_SCC(self, SCC, F):
                F_set.update(SCC)

        R_graph = self.get_reversed_graph()

        return R_graph.get_reachable_set_from(F_set)

    def label_fair_states(self, F):
        ''' Label all the fair states by a new atomic proposition.

        This method labels all the states from which a fair path exists by
        using a new atomic proposition that means "there exists a fair path
        from here". The new label is returned.

        :param F: a container of fairness constraints
        :type F: a container
        :returns: a new label that means "there exists a fair path from here"
        :rtype: str
        '''
        labels = self.labels()
        i = 0
        f_label = 'fair'
        while f_label in labels:
            f_label = 'fair{}'.format(i)
            i += 1

        for s in self.get_fair_states(F):
            self._labels[s].add(f_label)

        return f_label

    def __str__(self):
        ''' Return a string that represents a Kripke structure.

        :returns: a string that represents the Kripke
        :rtype: str
        '''
        return '(S={},S0={},R={},L={})'.format(self.states(),
                                               self.S0,
                                               list(self.transitions()),
                                               self._labels)



import random

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from trainify.utils import str_to_list


def create_modules(agent_config):
    """
    根据解析cfg文件的结果来构造模块
    :param blocks:
    :return:
    """
    blocks = agent_config.get("modules")
    assert blocks is not None, "没有给定actor网络参数"
    module = nn.Sequential()

    for index, x in enumerate(blocks[:]):

        # If it's a linear layer
        if (x["type"] == "linear"):
            # Get the info about the layer
            in_features = x["in_features"]
            out_features = x["out_features"]
            mean = x["mean"]
            std = x["std"]
            bias_zero = x["bias_zero"]
            activation = x["activation"]

            # Add the linear layer
            linear = nn.Linear(in_features=in_features, out_features=out_features)
            linear.weight.data.normal_(mean, std)
            if bias_zero:
                linear.bias.data.zero_()

            module.add_module("linear_{0}".format(index), linear)

            # activation
            if activation == "tanh":
                act = nn.Tanh()
                module.add_module("tanh_{0}".format(index), act)

    return module


class Actor(nn.Module):
    def __init__(self, agent_config):
        super(Actor, self).__init__()
        self.module_sequential = create_modules(agent_config)

    def forward(self, x):
        return self.module_sequential(x)


class Critic(nn.Module):
    def __init__(self, input_size, hidden_size, output_size):
        super().__init__()
        self.linear1 = nn.Linear(input_size, hidden_size)
        self.linear2 = nn.Linear(hidden_size, hidden_size)
        self.linear3 = nn.Linear(hidden_size, output_size)

    def forward(self, s, a):
        x = torch.cat([s, a], 1)
        x = F.relu(self.linear1(x))
        x = F.relu(self.linear2(x))
        x = self.linear3(x)
        return x


class DDPGAgent(object):
    def __init__(self, env, config):
        self.config = config
        self.originEnv = env
        self.env = env
        self.gamma = config['gamma']
        self.actor_lr = config['actor_lr']
        self.critic_lr = config['critic_lr']
        self.tau = config['tau']
        self.capacity = config['capacity']
        self.batch_size = config['batch_size']
        self.hidden_size = config['hidden_size']
        s_dim = self.env.observation_space.shape[0] * 2
        a_dim = self.env.action_space.shape[0]

        self.actor = Actor(self.config)
        self.network = Actor(self.config)
        self.actor_target = Actor(self.config)
        self.critic = Critic(s_dim + a_dim, self.hidden_size, a_dim)
        self.critic_target = Critic(s_dim + a_dim, self.hidden_size, a_dim)
        self.actor_optim = optim.Adam(self.actor.parameters(), lr=self.actor_lr)
        self.critic_optim = optim.Adam(self.critic.parameters(), lr=self.critic_lr)
        self.buffer = []

        self.actor_target.load_state_dict(self.actor.state_dict())
        self.critic_target.load_state_dict(self.critic.state_dict())

    def reset(self):
        self.env = self.originEnv
        self.gamma = self.config['gamma']
        self.actor_lr = self.config['actor_lr']
        self.critic_lr = self.config['critic_lr']
        self.tau = self.config['tau']
        self.capacity = self.config['capacity']
        self.batch_size = self.config['batch_size']

        s_dim = self.originEnv.observation_space.shape[0] * 2
        a_dim = self.originEnv.action_space.shape[0]

        self.actor = Actor(self.config)
        self.network = Actor(self.config)
        self.actor_target = Actor(self.config)
        self.critic = Critic(s_dim + a_dim, self.hidden_size, a_dim)
        self.critic_target = Critic(s_dim + a_dim, self.hidden_size, a_dim)
        self.actor_optim = optim.Adam(self.actor.parameters(), lr=self.actor_lr)
        self.critic_optim = optim.Adam(self.critic.parameters(), lr=self.critic_lr)
        self.buffer = []
        self.actor_target.load_state_dict(self.actor.state_dict())
        self.critic_target.load_state_dict(self.critic.state_dict())

    def act(self, s0):
        abs = str_to_list(s0)
        s0 = torch.tensor(abs, dtype=torch.float).unsqueeze(0)
        a0 = self.actor(s0).squeeze(0).detach().numpy()
        return a0

    def put(self, *transition):
        if len(self.buffer) == self.capacity:
            self.buffer.pop(0)
        self.buffer.append(transition)

    def learn(self):
        if len(self.buffer) < self.batch_size:
            return

        samples = random.sample(self.buffer, self.batch_size)

        s0, a0, r1, s1 = zip(*samples)

        s0 = torch.tensor(np.array(s0), dtype=torch.float)
        a0 = torch.tensor(np.array(a0), dtype=torch.float)
        r1 = torch.tensor(np.array(r1), dtype=torch.float).view(self.batch_size, -1)
        s1 = torch.tensor(np.array(s1), dtype=torch.float)

        def critic_learn():
            a1 = self.actor_target(s1).detach()
            y_true = r1 + self.gamma * self.critic_target(s1, a1).detach()

            y_pred = self.critic(s0, a0)

            loss_fn = nn.MSELoss()
            loss = loss_fn(y_pred, y_true)
            self.critic_optim.zero_grad()
            loss.backward()
            self.critic_optim.step()

        def actor_learn():
            loss = -torch.mean(self.critic(s0, self.actor(s0)))
            self.actor_optim.zero_grad()
            loss.backward()
            self.actor_optim.step()

        def soft_update(net_target, net, tau):
            for target_param, param in zip(net_target.parameters(), net.parameters()):
                target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

        critic_learn()
        actor_learn()
        soft_update(self.critic_target, self.critic, self.tau)
        soft_update(self.actor_target, self.actor, self.tau)


if __name__ == '__main__':
    agent_config = {
        "modules": [
            {
                "type": "linear",
                "in_features": 8,
                "out_features": 8,
                "mean": 0,
                "std": 0.1,
                "bias_zero": True,
                "activation": "tanh",
            },
            {
                "type": "linear",
                "in_features": 8,
                "out_features": 8,
                "mean": 0,
                "std": 0.1,
                "bias_zero": True,
                "activation": "tanh",
            },
            {
                "type": "linear",
                "in_features": 8,
                "out_features": 2,
                "mean": 0,
                "std": 0.1,
                "bias_zero": True,
                "activation": None,
            },
        ]
    }
    act = Actor(agent_config)
    print(act)



import time
import os
import re
import gym
import numpy as np

from trainify.abstract import initiate_divide_tool_rtree, initiate_divide_tool
from trainify.data import Recorder
from trainify.validator import cegar
from trainify.utils import str_to_list
from trainify.BBReach import do_BBReach
from trainify.env.verify import PendulumEnv

_DEFAULT_RESULT_PATH = os.path.join(os.path.expandvars('$HOME'), 'Trainify_data')
_DEFAULT_EXPERIMENT_NAME = "default_experiment_name"


class Trainify:
    def __init__(self,
                 env_config={},
                 env_class=None,
                 agent_config={},
                 agent_class=None,
                 verify_config={},
                 verify=False,
                 on_episode_end=None,
                 backend_channel=None,
                 # 输出路径配置
                 experiment_name=_DEFAULT_EXPERIMENT_NAME,
                 result_dir_name=None,  # 默认是实验名+时间
                 result_path=_DEFAULT_RESULT_PATH,  # 默认是用户主路径
                 ):
        # 导入配置
        self.env_config = env_config
        self.env_class = env_class
        self.agent_config = agent_config
        self.agent_class = agent_class
        self.verify_config = verify_config
        self.verify = verify
        self.backend_channel = backend_channel
        # 实验名称与输出路径设置，如果用户没有输入则使用默认值
        self.experiment_name = experiment_name
        self.time = time.strftime("_%Y%m%d_%H%M%S", time.localtime())
        self.experiment_name_with_time = experiment_name + self.time
        self.rtree_name = 'rtree_' + self.experiment_name_with_time
        self.result_path = result_path
        if result_dir_name is None:  # 默认是实验名+时间
            result_dir_name = self.experiment_name_with_time
        else:
            result_dir_name += self.time
        self.result_dir_name = result_dir_name

        # 将环境配置转成GYM格式
        self._np_state_space = np.array(self.env_config['state_space'])
        # Dict(x1:Box([-1.5], [1.5], (1,), float32), x2:Box([-1.5], [1.5], (1,), float32), x3:Box([-10.], [10.], (1,), float32))
        self._gym_dict_state_space = gym.spaces.Dict(self._handle_dict_env_state())
        # Box([ -1.5  -1.5 -10. ], [ 1.5  1.5 10. ], (3,), float32)
        self._gym_box_state_space = gym.spaces.Box(low=np.float32(self._np_state_space[0]),
                                                   high=np.float32(self._np_state_space[1]), dtype=np.float32)

        # 初始化各部件
        self.env = self.env_class()
        self.agent = self.agent_class(self.env, config=self.agent_config)
        self.recorder = Recorder(experiment_name=self.experiment_name,
                                 result_dir_name=self.result_dir_name,
                                 result_path=self.result_path,
                                 backend_channel=self.backend_channel)
        if self.verify:
            self.divide_tool = initiate_divide_tool_rtree(self.env_config['state_space'],
                                                          self.env_config['abs_initial_intervals'],
                                                          self.env_config['state_key_dim'],
                                                          self.rtree_name)
        else:
            self.divide_tool = initiate_divide_tool(self.env_config['state_space'],
                                                    self.env_config['abs_initial_intervals'])

        # 初始化附加函数
        self.save_model = self.recorder.create_save_model(self.agent, self.agent_config)
        self.load_model = self.recorder.create_load_model(self.agent, self.agent_config)
        self.recorder.save_model = self.save_model
        self.recorder.load_model = self.load_model
        self.agent.save_model = self.save_model
        self.agent.load_model = self.load_model
        self.create_cal_state_func(self.env_config, self.env)
        self.on_episode_end = on_episode_end

    def verify_cegar(self, verify_env=PendulumEnv, train_func=None):
        self.recorder.logger.info('Trainify 训练完毕 开始验证')
        if not self.verify:
            self.recorder.logger.info('Trainify 初始化Trainify时verify未设置为True，请重新初始化')
            return
        # TODO 构建验证用env
        if train_func is None: train_func = self.train_agent
        self.load_model()
        cegar(self.rtree_name,
              self.agent,
              self.divide_tool,
              train_func,
              verify_env(self.divide_tool, self.agent.actor),
              self.verify_config,
              self.recorder.logger)

    def train_agent(self, config={'step_num': 500, 'episode_num': 2000, 'reward_threshold': -3}, name=None):
        self.recorder.logger.info('Trainify 开始训练')
        step_num = config['step_num']
        episode_num = config['episode_num']
        reward_threshold = config['reward_threshold']
        if name is None: name = 'default_name' + time.strftime("_%Y%m%d_%H%M%S", time.localtime())
        self.recorder.create_experiment(name)
        for episode in range(episode_num):
            episode_reward = 0
            s = self.env.reset()
            abs = self.divide_tool.get_abstract_state(s)
            for step in range(step_num):
                # env.render()
                a = self.agent.act(abs)
                s_next, reward, done, _ = self.env.step(a)
                abs_next = self.divide_tool.get_abstract_state(s_next)
                self.agent.put(str_to_list(abs), a, reward, str_to_list(abs_next))
                episode_reward += reward
                self.agent.learn()
                s = s_next
                abs = abs_next
                if done: break
            if episode % 5 == 4:
                self.save_model()
            self.recorder.add_reward(episode_reward)
            if self.on_episode_end is not None:
                self.on_episode_end(episode, episode_reward)
            self.recorder.logger.warning('Trainify episode:' + str(episode) + ' episode_reward: ' + str(episode_reward))
            if episode >= 10 and np.min(self.recorder.get_reward_list()[-3:]) > reward_threshold:
                self.save_model()
                break
        self.save_model()
        self.recorder.logger.info('Trainify 训练结束')
        # self.verify_cegar()

        # self.recorder.writeAll2TensorBoard()

    def create_cal_state_func(self, config, Env):
        labels = [
            ['sin', 'math.sin'],
            ['cos', 'math.cos'],
            ['tan', 'math.tan'],
            ['tanh', 'math.tanh']
        ]

        def from_function(str, max):
            def cal(self, x):
                if max:
                    return eval(str)
                else:
                    return -eval(str)

            return cal

        for i, str in enumerate(config['dynamics']):
            for arr in labels:
                reg = re.compile(re.escape(arr[0]), re.IGNORECASE)
                config['dynamics'][i] = reg.sub(arr[1], config['dynamics'][i])

            setattr(Env, config['states_name'][i] + '_maximum',
                    from_function(config['dynamics'][i], True))
            setattr(Env, config['states_name'][i] + '_minimum',
                    from_function(config['dynamics'][i], False))
        self.recorder.logger.info('Trainify 在Env中加入状态最值计算函数成功')

    def _handle_dict_env_state(self):
        abs_obs = {}
        for i, name in enumerate(self.env_config['states_name']):
            abs_obs.update({
                name: gym.spaces.Box(
                    low=np.float32(self.env_config['state_space'][0][i]),
                    high=np.float32(self.env_config['state_space'][1][i]),
                    shape=(1,),
                    dtype=np.float32
                )
            })
        return abs_obs

    def _gen_dict(self, num):
        return {
            'min': np.full(shape=num, fill_value=float("inf"), dtype=float).tolist(),
            'max': np.full(shape=num, fill_value=-float("inf"), dtype=float).tolist()
        }

    def do_BBreach(self):
        self.load_model()
        do_BBReach(self.agent.actor, self.recorder, self.verify_config, self.env_config)


if __name__ == '__main__':
    # self.recorder.logger.info('main')
    env_c = {
        "dim": 2,
        "states_name": ["x1", "x2", "x3"],
        "state_space": [[-1.5, -1.5, -10], [1.5, 1.5, 10]],
        "abs_initial_intervals": [0.16, 0.16, 0.01],
        "state_key_dim": [0, 1],
        "dynamics": "x1^2+x2"
    }
    a = Trainify(env_config=env_c)



from pyModelChecking import Kripke
from pyModelChecking.CTLS import *

import unittest


class TestCTLSModelChecking(unittest.TestCase):
    def setUp(self):
        self.problems = [(Kripke(R=[(0, 0), (0, 1), (1, 2), (2, 2)],
                                 L={0: set([]),
                                    1: set(['p']),
                                    2: set(['q'])}),
                          [(Bool(True), set([0, 1, 2]), None),
                           (Bool(False), set([]), None),
                           (Not(True), set([]), None),
                           (A(F(G('q'))), set([1, 2]), None),
                           (A(Imply(F(G('q')), E(G('p')))), set([]), None)]),
                         (Kripke(R=[(0, 1), (0, 2), (1, 4), (4, 1), (4, 2),
                                    (2, 0), (3, 2), (3, 0), (3, 3), (6, 3),
                                    (2, 5), (5, 6)],
                                 L={0: set(),
                                    1: set(['Start', 'Error']),
                                    2: set(['Close']),
                                    3: set(['Close', 'Heat']),
                                    4: set(['Start', 'Close', 'Error']),
                                    5: set(['Start', 'Close']),
                                    6: set(['Start', 'Close', 'Heat'])}),
                          [(A(G(Imply(And(Not('Close'), 'Start'),
                                      A(Or(G(Not('Heat')), F(Not('Error'))))))
                              ), set([0, 1, 2, 3, 4, 5, 6]), None),
                           (A(G(Imply(And(Not('Close'), 'Start'),
                                      A(Or(G(Not('Heat')), F(Not('Error'))))))
                              ), set([]), set([6]))])]

    def test_modelchecking(self):
        for kripke, instances in self.problems:
            for formula, solution, Fconstraints in instances:
                S = modelcheck(kripke, formula, F=Fconstraints)

                self.assertEqual(set(S), solution)


if __name__ == '__main__':
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Trainify-proto",
    version="0.1.12",
    author="ericx",
    author_email="ericxlee@formail.com",
    description="Trainify-proto",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jieye-ericx/Trainify-proto",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'torch',
        'numpy',
        'scipy',
        'gym',
        'colorlog',
        'tensorboardx',
        'pandas',
        'rtree',
        'matplotlib',
        'lark',
        'setuptools'
    ]
)



import bisect
import copy
import time
from queue import Queue

import numpy as np
from rtree import index
from trainify.utils import str_to_list, list_to_str


def near_bound(current, target, standard):
    dim = len(current)
    half_dim = int(dim / 2)
    counter = 0
    record_dim = None
    for i in range(half_dim):
        if abs(current[i] - target[i]) > standard[i] or abs(current[i + half_dim] - target[i + half_dim]) > \
                standard[i]:
            counter += 1
            record_dim = i
    if counter <= 0 or contain(current, target):
        return True
    elif counter == 1:
        if current[record_dim] - target[record_dim + half_dim] <= standard[record_dim] or target[record_dim] - \
                current[
                    record_dim + half_dim] <= standard[record_dim]:
            return True
        elif (target[record_dim] < current[record_dim] < target[record_dim + half_dim]) or (
                current[record_dim] < target[record_dim] < current[record_dim + half_dim]):
            return True
        else:
            return False
    else:
        return False


def combine_bound_list(bound_list, std):
    relation = []
    length = len(bound_list)
    compacted_bound_list = []
    for i in range(length):
        relation.append([False] * length)
    for i in range(length):
        for j in range(length):
            if i != j and near_bound(bound_list[i], bound_list[j], std):
                relation[i][j] = True
                relation[j][i] = True
    # 先构造邻接链表，之后根据表格合并bound
    flag_list = [False for i in range(length)]
    for i in range(length):
        if flag_list[i]:
            continue
        bfs = Queue()
        bfs_set = set()
        bfs.put(i)
        tmp = bound_list[i]
        while not bfs.empty():
            index = bfs.get()
            flag_list[index] = True
            tmp = combine(tmp, bound_list[index])
            bfs_set.add(index)
            near_list = get_near_bound_list(relation, index, bound_list)
            for near_index in near_list:
                if near_index not in bfs_set:
                    bfs.put(near_index)
        compacted_bound_list.append(tmp)

    return compacted_bound_list


# 如果bound的范围大于diameter，则进行二分
def split(bound_list, diameter):
    assert len(bound_list) > 0
    dim = len(bound_list[0])
    half_dim = int(dim / 2)
    assert half_dim == len(diameter)
    res_list = []
    for bound in bound_list:
        split_list = [bound]
        for i in range(half_dim):
            if bound[i + half_dim] - bound[i] > diameter[i]:
                tmp_list = []
                for b in split_list:
                    b1 = copy.deepcopy(b)
                    b2 = copy.deepcopy(b)
                    mid = (bound[i] + bound[i + half_dim]) / 2
                    b1[i] = mid
                    b2[i + half_dim] = mid
                    tmp_list.append(b1)
                    tmp_list.append(b2)
                split_list = copy.deepcopy(tmp_list)
        res_list = res_list + split_list
    return res_list


def contain(current, target):
    dim = len(current)
    half_dim = int(dim / 2)
    cur_in_tar = True
    tar_in_cur = True
    for i in range(half_dim):
        if not (current[i] >= target[i] and current[i + half_dim] <= target[i + half_dim]):
            cur_in_tar = False
        if not (target[i] >= current[i] and target[i + half_dim] <= current[i + half_dim]):
            tar_in_cur = False
    return cur_in_tar or tar_in_cur


# 获取与指定index的bound能合并的bound list
def get_near_bound_list(relation, current_index, bound_list):
    res_list = []
    for i in range(len(bound_list)):
        if i != bound_list and relation[current_index][i]:
            res_list.append(i)
    return res_list


def combine(current, target):
    res = []
    dim = len(current)
    half_dim = int(dim / 2)
    for i in range(dim):
        if i < half_dim:
            res.append(min(current[i], target[i]))
        else:
            res.append(max(current[i], target[i]))
    return res


def max_min_clip(bound, state):
    dim = len(bound)
    half_dim = int(dim / 2)
    s = copy.deepcopy(state)
    for i in range(half_dim):
        s[i] = max(s[i], bound[i])
        s[i + half_dim] = min(s[i + half_dim], bound[i + half_dim])
    return s


# 用于生成不带rtree的divide_tool
def initiate_divide_tool(state_space, initial_intervals):
    divide_point = []
    dp_sets = []
    for i in range(len(state_space[0])):
        lb = state_space[0][i]
        ub = state_space[1][i]
        tmp = [lb]
        dp_set = set()
        while lb < ub:
            lb = round(lb + initial_intervals[i], 10)
            lb = min(lb, state_space[1][i])
            tmp.append(lb)
            dp_set.add(lb)
        divide_point.append(tmp)
        dp_sets.append(dp_set)
    return DivideTool(divide_point, dp_sets)


def initiate_adaptive_divide_tool(state_space, initial_intervals, key_dim, file_name):
    # divide_point = []
    # dp_sets = []
    # for i in range(len(state_space[0])):
    #     lb = state_space[0][i]
    #     ub = state_space[1][i]
    #     tmp = [lb]
    #     dp_set = set()
    #     dp_set.add(lb)
    #     while lb < ub:
    #         lb = round(lb + initial_intervals[i], 10)
    #         tmp.append(lb)
    #         dp_set.add(lb)
    #     divide_point.append(tmp)
    #     dp_sets.append(dp_set)
    small_initial_interval = [division(num) for num in initial_intervals]
    fine_grained_divide_tool = initiate_divide_tool(state_space, small_initial_interval)
    normal_grained_divide_tool = initiate_divide_tool(state_space, initial_intervals)
    adp = initiate_divide_tool_rtree(state_space, initial_intervals, key_dim, file_name)
    adp.fine_grained_divide_tool = fine_grained_divide_tool
    adp.fine_grained_interval = small_initial_interval
    adp.normal_grained_divide_tool = normal_grained_divide_tool
    if adp.rtree is None:
        return adp, 0
    return adp, adp.rtree.get_size()


# 用于生成带有rtree的divide_tool
def initiate_divide_tool_rtree(state_space, initial_intervals, key_dim, file_name):
    divide_point = []
    dp_sets = []
    # for i in range(len(state_space[0])):
    #     lb = state_space[0][i]
    #     ub = state_space[1][i]
    #     tmp = [lb]
    #     dp_set = set()
    #     dp_set.add(lb)
    #     while lb < ub:
    #         lb = round(lb + initial_intervals[i], 10)
    #         tmp.append(lb)
    #         dp_set.add(lb)
    #     dp_sets.append(dp_set)

    key_state_space = [[], []]
    key_initial_intevals = []
    for i in range(len(state_space[0])):
        if key_dim is not None:
            if i in key_dim:
                key_state_space[0].append(state_space[0][i])
                key_state_space[1].append(state_space[1][i])
                key_initial_intevals.append(initial_intervals[i])
                continue
        lb = state_space[0][i]
        ub = state_space[1][i]
        tmp = [lb]
        dp_set = set()
        dp_set.add(lb)
        while lb < ub:
            lb = round(lb + initial_intervals[i], 10)
            tmp.append(lb)
            dp_set.add(lb)
        dp_sets.append(dp_set)
        divide_point.append(tmp)
    rtree = None
    if key_dim is not None:
        p = index.Property()
        p.dimension = len(key_dim)
        # rtree = index.Index(file_name, properties=p)
        # if rtree.get_size() == 0:
        rtree = index.Index(file_name, divide(key_state_space, key_initial_intevals), properties=p)
        print('DivideTool rtree状态数量', rtree.get_size())
    else:
        key_dim = []
    divide_tool = DivideTool(divide_point, dp_sets)
    divide_tool.key_dim = key_dim
    divide_tool.rtree = rtree
    return divide_tool


# 用于yield rtree构造中的状态的上下界
def divide(state_space, intervals):
    lb = state_space[0]
    ub = state_space[1]
    np_initial_intervals = np.array(intervals)
    id = 1
    while True:
        flag = True
        for j in range(len(lb)):
            if lb[j] >= ub[j]:
                flag = False
                break
        if flag:
            upper = np.array(lb) + np_initial_intervals
            for i in range(len(upper)):
                upper[i] = round(upper[i], 10)
            # print(str(lb) + str(upper))
            # print(id)
            obj_str = ','.join([str(_) for _ in lb]) + ',' + ','.join([str(_) for _ in upper])
            # print(obj_str)
            yield id, tuple(lb) + tuple(upper.tolist()), obj_str
            # rtree.insert(id, tuple(lb) + tuple(upper.tolist()),
            #              obj=obj_str)
            id += 1
            if id % 1000 == 0:
                print(id)
        i, lb = get_next_lb(lb, state_space, intervals)
        # print(lb)
        if lb is None:
            break


# 给定一个当前的下界lb，以及状态范围和划分粒度，返回下一个lb
def get_next_lb(lb, state_space, intervals):
    tmp_lb = copy.deepcopy(lb)
    flag = False
    for i in range(len(tmp_lb)):
        if tmp_lb[i] < state_space[1][i]:
            flag = True
            tmp_lb[i] = round(intervals[i] + tmp_lb[i], 10)
            if i >= 1:
                j = i - 1
                while j >= 0:
                    tmp_lb[j] = state_space[0][j]
                    j -= 1
            # if i >= 2:
            #     tmp_lb[1] = state_space.lb[1]
            # if i >= 3:
            #     tmp_lb[2] = state_space.lb[2]
            return i, tmp_lb
    if not flag:
        return -1, None


# 将一个bound的各个维度进行二分
def bound_bisect(bound):
    dim = int(len(bound) / 2)
    bounds = []
    interval = []
    state_space = [[], []]
    for i in range(dim):
        interval.append(round((bound[i + dim] - bound[i]) / 2, 10))
        state_space[0].append(bound[i])
        state_space[1].append(bound[i + dim])
    lb = bound[0:dim]

    count = 0
    # print(bound)
    while lb is not None:
        upper = np.array(lb) + np.array(interval)
        flag = True
        for i in range(len(upper)):
            upper[i] = round(upper[i], 10)
            if upper[i] > bound[i + dim]:
                flag = False
                break
        if flag:
            bounds.append(lb + upper.tolist())
        # print(lb)
        id1, lb = get_next_lb(lb, state_space, interval)
        count += 1
    return bounds


def division(num):
    granularity = 10
    return round(num / granularity, 10)


class DivideTool:
    def __init__(self, divide_point, dp_sets):
        self.divide_point = divide_point
        self.dp_sets = dp_sets
        self.key_dim = []
        self.rtree = None
        self.fine_grained_divide_tool = None
        self.fine_grained_interval = None
        self.normal_grained_divide_tool = None
        self.restore_dp = copy.deepcopy(divide_point)
        self.restore_dp_sets = copy.deepcopy(dp_sets)
        # self.part_state1 = None
        # self.part_state2 = None

    # 将给定的状态范围，分成普通的状态空间和关键状态空间
    def part_state(self, bound):
        dim = len(bound)
        half_dim = int(dim / 2)
        state_space = list(range(dim - 2 * len(self.key_dim)))
        key_state_space = list(range(2 * len(self.key_dim)))
        half_dim1 = int(len(state_space) / 2)
        half_dim2 = int(len(key_state_space) / 2)
        id1 = 0
        id2 = 0
        for i in range(half_dim):
            if i in self.key_dim:
                key_state_space[id2] = bound[i]
                key_state_space[id2 + half_dim2] = bound[i + half_dim]
                id2 += 1
            else:
                state_space[id1] = bound[i]
                state_space[id1 + half_dim1] = bound[i + half_dim]
                id1 += 1
        return state_space, key_state_space

    # 给定范围查询
    def intersection(self, bound):
        dim = len(bound)
        half_dim = int(dim / 2)
        state_space, key_state_space = self.part_state(bound)
        half_dim1 = int(len(state_space) / 2)
        half_dim2 = int(len(key_state_space) / 2)

        # 获取各个维度的划分点
        t0 = time.time()
        region = self.get_abstract_region(state_space)
        t1 = time.time()
        state_list = []
        abstract_state = list(range(len(state_space)))
        self.point_to_state(region, 0, abstract_state, state_list)
        t2 = time.time()
        # print((t2 - t1) / (t1 - t0))
        length1 = len(state_list)

        res_list = []
        # 关键维度上的查询
        if self.rtree is not None:
            rtree_state_list = list(self.rtree.intersection(key_state_space, objects=True))
            length2 = len(rtree_state_list)

            final_state_list = []
            # for i in range(len(state_space)):

            for i in range(length1):
                part_state1 = state_list[i]
                for j in range(length2):
                    ik1 = 0
                    ik2 = 0
                    final_abs = list(range(dim))
                    part_state2 = rtree_state_list[j]
                    for k in range(half_dim):
                        if k in self.key_dim:
                            final_abs[k] = part_state2.bbox[ik2]
                            final_abs[k + half_dim] = part_state2.bbox[ik2 + half_dim2]
                            ik2 += 1
                        else:
                            final_abs[k] = part_state1[ik1]
                            final_abs[k + half_dim] = part_state1[ik1 + half_dim1]
                            ik1 += 1
                    final_state_list.append(final_abs)
            res_list = list_to_str(final_state_list)
        else:
            res_list = list_to_str(state_list)
        return res_list

    # 给定一个范围，返回在相应维度上的分界点
    def get_abstract_region(self, s):
        # s = s.tolist()
        dim = int(len(s) / 2)
        tmp = []

        # 范围查询
        for i in range(dim):
            tmp2 = []
            pos_left = bisect.bisect_left(self.divide_point[i], s[i])
            pos_right = bisect.bisect(self.divide_point[i], s[i + dim])
            if s[i] != self.divide_point[i][pos_left]:
                pos_left -= 1
            while pos_left <= pos_right and pos_left < len(self.divide_point[i]):
                try:
                    tmp2.append(self.divide_point[i][pos_left])
                except:
                    print('debug')
                pos_left += 1
            tmp.append(tmp2)
            # tmp[i] = divide_point[i][pos_left - 1]
            # tmp[i + dim] = divide_point[i][pos_right]
        return tmp

    # 具体状态到抽象状态，即查询具体状态
    def get_abstract_state(self, s):
        dim = len(s)
        # s需要是list
        # s = s.tolist()
        if type(s) is np.ndarray:
            s = s.tolist()
        bound = s + s
        # s1, s2 = self.part_state(bound)
        tt = self.intersection(bound)
        if len(tt) == 0:
            print('bug')
        return tt[0]
        # tmp = []
        # # 点查询
        # if dim == len(self.divide_point):
        #     tmp = list(range(2 * dim))
        #     for i in range(dim):
        #         pos = bisect.bisect(self.divide_point[i], s[i])
        #         tmp[i] = self.divide_point[i][pos - 1]
        #         tmp[i + dim] = self.divide_point[i][pos]
        #     obj_str = ','.join([str(_) for _ in tmp])
        #     return obj_str
        # return None

    # 将划分边界点，转化为抽象状态list
    def point_to_state(self, divide_point, current_dim, abstract_state, state_list):
        dim = len(divide_point)
        # abstract_state = list(range(2 * dim))

        if current_dim < dim:
            if len(divide_point[current_dim]) == 1:
                a_s = copy.deepcopy(abstract_state)
                a_s[current_dim] = divide_point[current_dim][0]
                a_s[current_dim + dim] = divide_point[current_dim][0]
                self.point_to_state(divide_point, current_dim + 1, a_s, state_list)
            else:
                for i in range(len(divide_point[current_dim]) - 1):
                    a_s = copy.deepcopy(abstract_state)
                    a_s[current_dim] = divide_point[current_dim][i]
                    try:
                        a_s[current_dim + dim] = divide_point[current_dim][i + 1]
                    except:
                        print('ssss')
                    self.point_to_state(divide_point, current_dim + 1, a_s, state_list)
        else:
            state_list.append(abstract_state)

    def rtree_refinement(self, violate_states, start_id):
        print(self.rtree.bounds)
        all_states = list(self.rtree.intersection(self.rtree.bounds, objects=True))
        print('所有状态数量', len(all_states))
        key_dim = len(self.key_dim)
        cnt1 = 0
        cnt2 = 0
        cnt3 = 0
        set_vio = set(violate_states)
        for state in all_states:
            state_str = state.object
            state_str = ','.join([str(_) for _ in state.bbox])
            if state_str in set_vio:
                cnt1 += 1
                if cnt1 % 1000 == 0:
                    print('cnt1', cnt1)
                refine_bounds = bound_bisect(str_to_list(state_str))
                for b in refine_bounds:
                    obj_str = ','.join([str(_) for _ in b[0:key_dim]]) + ',' + ','.join(
                        [str(_) for _ in b[key_dim:2 * key_dim]])
                    # print(obj_str)
                    # rtree.insert(start_id, tuple(b), obj=obj_str)
                    # print(start_id)
                    cnt3 += 1
                    yield start_id, tuple(b), obj_str
                    start_id += 1
            else:
                cnt2 += 1
                yield state.id, tuple(state.bbox), state_str
        print('---------', cnt1, cnt2, cnt3)
        return start_id

    def adaptive_partition(self, abstract_state, bound, max_id):
        # bound_state_list = self.intersection(bound)
        if isinstance(abstract_state, str):
            abstract_state = str_to_list(abstract_state)
        old_key_states = [tuple(abstract_state)]
        # new_key_states = [tuple(abstract_state)]
        new_key_states = []
        dim = len(bound)
        assert dim == len(abstract_state)
        half_dim = int(dim / 2)
        key_dim_num = 0

        n, k = self.part_state(abstract_state)
        a_obj = []
        if len(k) != 0:
            a_obj = [val for val in list(self.rtree.intersection(tuple(k), objects=True)) if val.bbox == k]
        deal_key_dim = len(a_obj) != 0

        for i in range(half_dim):
            if i in self.key_dim:
                key_dim_num += 1
            # 抽象状态的下界低于了bound
            if abstract_state[i] < round(bound[i] - self.fine_grained_interval[i], 10):
                # 关键维度，需要在rtree里处理
                if i in self.key_dim and deal_key_dim:
                    new_key_states = self.get_new_key_states(i, half_dim, bound[i], old_key_states)
                    old_key_states = copy.deepcopy(new_key_states)
                    continue
                else:
                    self.add_divide_line(i - key_dim_num, bound[i])
            # 抽象状态的上界大于了bound
            if abstract_state[i + half_dim] > round(bound[i + half_dim] + self.fine_grained_interval[i], 10):
                # 关键维度，需要在rtree里处理
                if i in self.key_dim and deal_key_dim:
                    new_key_states = self.get_new_key_states(i, half_dim, bound[i + half_dim],
                                                             old_key_states, False)
                    old_key_states = copy.deepcopy(new_key_states)
                else:
                    self.add_divide_line(i - key_dim_num, bound[i + half_dim], False)

        if deal_key_dim and len(new_key_states) != 0:
            self.rtree.delete(a_obj[0].id, tuple(k))
            for new_ks in new_key_states:
                a, b = self.part_state(new_ks)
                obj_str = ','.join([str(_) for _ in b])
                self.rtree.insert(max_id + 1, tuple(b), obj_str)
                max_id += 1
        return max_id

    # 在非关键维度上增加分界线
    def add_divide_line(self, dim, value, low=True):
        if low:
            p1 = bisect.bisect_left(self.fine_grained_divide_tool.divide_point[dim], value)
            assert p1 > 0
            # 查找分割线
            element = self.fine_grained_divide_tool.divide_point[dim][p1 - 1]
            if element not in self.dp_sets[dim]:
                bisect.insort(self.divide_point[dim], element)
                self.dp_sets[dim].add(element)
            # if self.fine_grained_divide_tool.divide_point[dim][p1 - 1] not in self.divide_point[dim]:
            #     bisect.insort(self.divide_point[dim], self.fine_grained_divide_tool.divide_point[dim][p1 - 1])
            # bisect.insort(self.divide_point[dim], self.fine_grained_divide_tool.divide_point[dim][p1 - 1])
        else:
            p2 = bisect.bisect(self.fine_grained_divide_tool.divide_point[dim], value)
            element = self.fine_grained_divide_tool.divide_point[dim][p2]
            if element not in self.dp_sets[dim]:
                bisect.insort(self.divide_point[dim], element)
                self.dp_sets[dim].add(element)
            # if self.fine_grained_divide_tool.divide_point[dim][p2] not in self.divide_point[dim]:
            #     bisect.insort(self.divide_point[dim], self.fine_grained_divide_tool.divide_point[dim][p2])
            # bisect.insort(self.divide_point[dim], self.fine_grained_divide_tool.divide_point[dim][p2])

    # 对抽象状态进行rtree维度上的划分
    def get_new_key_states(self, dim, half_dim, value, old_key_states, low=True):
        new_key_states = []
        if low:
            p1 = bisect.bisect_left(self.fine_grained_divide_tool.divide_point[dim], value)
            assert p1 > 0
            element = self.fine_grained_divide_tool.divide_point[dim][p1 - 1]
        else:
            p2 = bisect.bisect(self.fine_grained_divide_tool.divide_point[dim], value)
            element = self.fine_grained_divide_tool.divide_point[dim][p2]
        for old_key_state in old_key_states:
            new_key_state1 = list(old_key_state)
            new_key_state2 = list(old_key_state)

            new_key_state1[dim + half_dim] = element
            new_key_state2[dim] = element

            new_key_states.append(tuple(new_key_state1))
            new_key_states.append(tuple(new_key_state2))
        return new_key_states

    def restore(self):
        self.divide_point = copy.deepcopy(self.restore_dp)
        self.dp_sets = copy.deepcopy(self.restore_dp_sets)


class AdaptiveDivideTool(DivideTool):
    def __init__(self, divide_point, dp_sets, fine_grained_divide_tool, fine_grained_interval, key_dim=None,
                 file_name=None):
        super().__init__(divide_point, dp_sets)
        self.fine_grained_divide_tool = fine_grained_divide_tool
        self.fine_grained_interval = fine_grained_interval
        self.key_dim = key_dim


if __name__ == "__main__":
    print('test')
    # dt = DivideTool([[-1, 0, 1], [-0.5, 0, 0.5]])
    # abstract_state = list(range(4))
    # state_list = []
    # dt.point_to_state([[-1, 0], [-0.5, 0, 0.5]], 0, abstract_state, state_list)
    # res = dt.intersection([0, -0.3, 0, 0.1])
    # print('eeee')

    # sp = split([[0, 0, 1, 1], [0, 0, 1.5, 1.5], [0, 0, 0.5, 2]], [1, 1])
    #
    # initial_interval = [0.1, 0.01, 0.1, 0.2]
    # lb = [-1, 0, -1, -2]
    # ub = [-0.5, 2, -0.5, 2]
    # adt = initiate_adaptive_divide_tool([lb, ub], initial_interval, [0, 1, 2, 3], 'adt')
    # max_id = adt.adaptive_partition([-0.9, 0.01, -0.9, -1.8, -0.8, 0.02, -0.8, -1.6],
    #                                 [-0.82, 0.015, -1, -2, -0.7, 0.03, -0.87, -1.64], 1200)
    # res = adt.intersection([-0.82, 0.015, -1, -2, -0.7, 0.03, -0.87, -1.61])
    #
    # small_initial_interval = [division(num) for num in initial_interval]
    #
    # dt = initiate_divide_tool_rtree([lb, ub], initial_interval, [1, 2], 'divide_tool_test1')
    #
    # # res = dt.intersection([-0.7, 0.2, -0.9, 0, -0.6, 0.3, -0.6, 9])
    # res2 = dt.get_abstract_state([-0.6, 1, -0.8, 1])
    # vio_states = ['1.01,-0.8,0.2,1.02,-0.7,0.4']
    #
    # p = index.Property()
    # p.dimension = len(dt.key_dim)
    # dt.rtree = index.Index('divide_tool_test1', dt.rtree_refinement(vio_states, dt.rtree.get_size()), properties=p)
    # print('size:', dt.rtree.get_size())
    # res3 = dt.intersection([-0.65, 1.01, -0.8, 0.2, -0.65, 1.02, -0.7, 0.4])
    # print('ssss')

    # state_space = [[0, -0.5], [100, 0.5]]
    # divide_tool = initiate_divide_tool(state_space, [0.1,0.01])
    # print(divide_tool.divide_point)
    # print(divide_tool.dp_sets)




from gym.utils import seeding

from os import path


import math
import gym
from gym import spaces, logger
from gym.utils import seeding
import numpy as np
# from gym.envs.classic_control import rendering


class Cartpole(gym.Env):

    metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second': 30
    }

    def __init__(self):
        self.gravity = 9.8  # 重力加速度g
        self.masscart = 1.0  # 车重量
        self.masspole = 0.1  # 杆重量
        self.total_mass = (self.masspole + self.masscart)  # （0.1 + 1） = 1.1，车和杆的重量
        self.length = 0.5  # 计算杆高度用
        self.polemass_length = (self.masspole * self.length)  # （0.1 * 0.5） = 0.05，计算加速度用
        self.force_mag = 10.0  # 输入动作，每次施加给车的力
        self.tau = 0.02  # 状态state更新之间的秒数
        self.kinematics_integrator = 'euler'  # 运动学积分器，为了计算下一步state

        # Angle at which to fail the episode 本局失败的角度
        self.theta_threshold_radians = 12 * 2 * math.pi / 360  # 12 * 2 * 180° / 360° = 12 °
        self.x_threshold = 2.4

        # Angle limit set to 2 * theta_threshold_radians so failing observation
        # 角度限制设为 2 * theta_threshold_radians，这就是失败的观察

        # is still within bounds. 仍在范围内
        high = np.array([self.x_threshold * 2,  # 4.8
                         np.finfo(np.float32).max,  # 取float的最大值
                         self.theta_threshold_radians * 2,  # 24°
                         np.finfo(np.float32).max],  # 取float的最大值
                        dtype=np.float32)

        self.action_space = spaces.Discrete(2)  # 离散动作定义（2个动作）为0,1

        self.observation_space = spaces.Box(-high, high, dtype=np.float32)  # 连续状态定义(四种都是连续的)
        # Num     Observation               Min                     Max
        # 0       Cart Position             -4.8                    4.8
        # 1       Cart Velocity             -Inf                    Inf
        # 2       Pole Angle                -0.418 rad (-24 deg)    0.418 rad (24 deg)
        # 3       Pole Angular Velocity     -Inf                    Inf

        self.seed()
        self.viewer = None
        self.state = None
        self.steps_beyond_done = None


    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self,  action):

        err_msg = "%r (%s) invalid" % (action, type(action))
        assert self.action_space.contains(action), err_msg  # 判断动作空间是否包含输入的action，否则报err_msg的错

        x, x_dot, theta, theta_dot = self.state  # 系统当前状态，包括车位置x,车速x_dot, 杆角度theta, 杆顶端的速度theta_dot
        force = self.force_mag if action == 1 else -self.force_mag  # 输入动作，（1代表右），即作用到车上的力
        costheta = math.cos(theta)  # 计算角度cos值
        sintheta = math.sin(theta)  # 计算角度sin值

        # 施加力对杆子和小车影响的数学公式
        # temp:车摆的动力学方程式，即加速度与动作之间的关系
        temp = (force + self.polemass_length * theta_dot ** 2 * sintheta) / self.total_mass
        # temp=(±10 + 0.05 * 杆顶速度^2 * sin() ) / 1.1
        # thetaacc:摆的角加速度
        thetaacc = (self.gravity * sintheta - costheta * temp) / (
                self.length * (4.0 / 3.0 - self.masspole * costheta ** 2 / self.total_mass))
        # 小车的平加速度
        xacc = temp - self.polemass_length * thetaacc * costheta / self.total_mass

        # 计算下一步的state（x, x_dot, theta, theta_dot）
        if self.kinematics_integrator == 'euler':  # （欧拉）
            # tau是更新步长0.02
            x = x + self.tau * x_dot
            x_dot = x_dot + self.tau * xacc
            theta = theta + self.tau * theta_dot
            theta_dot = theta_dot + self.tau * thetaacc
        else:  # 半隐式欧拉
            x_dot = x_dot + self.tau * xacc
            x = x + self.tau * x_dot
            theta_dot = theta_dot + self.tau * thetaacc
            theta = theta + self.tau * theta_dot

        self.state = (x, x_dot, theta, theta_dot)
        # state包括车位置x,车速x_dot, 杆角度theta, 杆顶端的速度theta_dot

        # 结束的条件
        done = bool(
            x < -self.x_threshold
            or x > self.x_threshold  # 车的位置小于-2.4，大于2.4
            or theta < -self.theta_threshold_radians
            or theta > self.theta_threshold_radians  # 角度小于-12°，大于12°
        )

        # 奖励
        if not done:
            reward = 1.0
        elif self.steps_beyond_done is None:
            # Pole just fell! 杆刚落下
            self.steps_beyond_done = 0
            reward = 1.0
        else:
            if self.steps_beyond_done == 0:
                logger.warn("您正在调用‘step（）’，即使此环境已返回done=True。"
                            "一旦收到‘done=True’，您应该始终调用‘reset（）’——任何进一步的步骤都是未定义的行为。")

            self.steps_beyond_done += 1  # 当前的step加一
            reward = 0.0  # 奖励为0

        return np.array(self.state), reward, done, {}


    def reset(self,):
        pass

    def _get_obs(self):
        pass

    def render(self, mode="human"):

        screen_width = 600
        screen_heigt = 400  # 画布的高宽

        world_width = self.x_threshold * 2  # 整个小车在x轴的位置范围（2.4*2 = 4.8）
        scale = screen_width / world_width  # 尺寸600/4.8 =125
        carty = 100  # 小车离画布低端100
        polewidth = 10.0  # 杆的宽度
        polelen = scale * (2 * self.length)  # 125*（2*0.5）=125
        cartwidth = 50.0  # 小车宽度
        cartheight = 30.0  # 小车高读

        if self.viewer is None:  # 如果显示器是空的
            from gym.envs.classic_control import rendering
            self.viewer = rendering.Viewer(screen_width, screen_heigt)  # 显示画布

            # 创建车
            l, r, t, b = -cartwidth / 2, cartwidth / 2, cartheight / 2, -cartheight / 2
            # l = -25, r = 25, t = 15, b = -15
            axleoffset = cartheight / 4.0  # 车轴7.5
            cart = rendering.FilledPolygon([(l, b), (l, t), (r, t), (r, b)])
            # 通过长方形的四个点来画小车，以小车的中心为原点，顺时针对应四个点的顺序（从第三象限为第一个点）
            # Transform给cart添加平移属性和旋转属性
            self.carttrans = rendering.Transform()
            cart.add_attr(self.carttrans)
            self.viewer.add_geom(cart)  # 在显示器上添加小车

            # 创建杆
            l, r, t, b = -polewidth / 2, polewidth / 2, polelen - polewidth / 2, -polewidth / 2
            # l = -5, r = 5, t =120，b=-5
            pole = rendering.FilledPolygon([(l, b), (l, t), (r, t), (r, b)])
            # 通过长方形的四个点来画杆，但是是以小车的中心为原点
            pole.set_color(.8, .6, .4)  # 给杆设置颜色
            # 添加摆杆装换矩阵属性
            self.poletrans = rendering.Transform(translation=(0, axleoffset))  # 杆中心平移（）
            pole.add_attr(self.poletrans)  # 对杆进行平移赋值
            pole.add_attr(self.carttrans)  # 对杆进行平移赋值
            self.viewer.add_geom(pole)  # 在显示器上添加杆

            # 创建摆杆和车之间的连接
            self.axle = rendering.make_circle(polewidth / 2)  # 圆的直径为5
            self.axle.add_attr(self.poletrans)  # 对圆进行平移赋值
            self.axle.add_attr(self.carttrans)  # 对圆进行平移赋值
            self.axle.set_color(.5, .5, .8)  # 给圆设置颜色
            self.viewer.add_geom(self.axle)  # 在显示器上添加圆

            # 创建轨道，即一条线
            self.track = rendering.Line((0, carty), (screen_width, carty))  # (0,100),(600,100)
            self.track.set_color(0, 0, 0)  # 给线设置颜色
            self.viewer.add_geom(self.track)  # 在显示器上添加线

            self._pole_geom = pole  # ???

        if self.state is None:
            return None

        # 编辑极点多边形顶点
        pole = self._pole_geom
        l, r, t, b = -polewidth / 2, polewidth / 2, polelen - polewidth / 2, -polewidth / 2
        # l = -5, r = 5, t = 120, b = -5
        pole.v = [(l, b), (l, t), (r, t), (r, b)]

        # 设置平移属性
        x = self.state  # 这里的x是状态，那么状态有四个属性，包括车位置x,车速x_dot, 杆角度theta, 杆顶端的速度theta_dot
        cartx = x[0] * scale + screen_width / 2.0  # 车中央  车位置 *125 +600/2
        self.carttrans.set_translation(cartx, carty)  # ？？？
        self.poletrans.set_rotation(-x[2])  # 杆角度的负数

        return self.viewer.render(return_rgb_array=mode == 'rgb_array')


    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None


from pyModelChecking.graph import *
import unittest


class TestDiGraph(unittest.TestCase):

    def setUp(self):
        self.V = set([0, 1, 3, 4])
        self.E = set([(0, 2), (2, 2), (0, 1), (1, 0)])

        self.G = DiGraph(self.V, self.E)

    def test_init(self):
        V = set(self.V)
        for (s, d) in self.E:
            V.add(s)
            V.add(d)

        self.assertEqual(set(self.G.nodes()), V)

    def test_nodes(self):
        V = (self.V | set([2]))

        self.assertEqual(set(self.G.nodes()), V)

    def test_edges(self):
        self.assertEqual(set(self.G.edges_iter()), self.E)

    def test_add_node(self):

        V = (set(self.G.nodes()) | set([5]))

        self.G.add_node(5)

        self.assertEqual(set(self.G.nodes()), V)

        with self.assertRaises(RuntimeError):
            self.G.add_node(0)

    def test_add_edge(self):

        E = set(self.G.edges_iter()) | set([(0, 3)])
        V = set(self.G.nodes()) | set([0, 3])

        self.G.add_edge(0, 3)

        self.assertEqual(set(self.G.nodes()), V)
        self.assertEqual(set(self.G.edges_iter()), E)

        self.G.add_edge(0, 6)

        E = E | set([(0, 6)])
        V = V | set([0, 6])

        self.assertEqual(set(self.G.nodes()), V)
        self.assertEqual(set(self.G.edges_iter()), E)

        with self.assertRaises(RuntimeError):
            self.G.add_edge(0, 6)

    def test_sources(self):
        S = set()
        for (s, d) in self.G.edges_iter():
            S.add(s)

        self.assertEqual(set(self.G.sources()), S)

    def test_next(self):

        for s in set(self.G.nodes()):
            N = set()
            for (p, q) in self.G.edges_iter():
                if s == p:
                    N.add(q)

            self.assertEqual(self.G.next(s), N)

    def test_reversed(self):
        rG = self.G.get_reversed_graph()

        rE = []
        for (s, d) in self.G.edges_iter():
            rE.append((d, s))

        self.assertEqual(set(rE), set(rG.edges_iter()))
        self.assertEqual(set(self.G.nodes()), set(rG.nodes()))

    def test_subgraph(self):
        Vp = set([0, 1, 4, 5])
        SG = self.G.get_subgraph(Vp)

        GE = set(self.G.edges_iter())
        self.assertEqual(Vp & set(self.G.nodes()), set(SG.nodes()))
        for (s, d) in SG.edges_iter():
            self.assertIn(s, Vp)
            self.assertIn(d, Vp)

            self.assertIn((s, d), GE)

    def test_strongly_connected_components(self):
        SCCs = set([frozenset([0, 1]), frozenset([2]), frozenset([3]),
                    frozenset([4])])

        list_SCCs = [frozenset(s) for s in compute_SCCs(self.G)]
        computed_SCCs = set(list_SCCs)

        self.assertEqual(computed_SCCs, SCCs)


if __name__ == '__main__':
    unittest.main()



"""
.. module:: graph
   :synopsis: A module to represent directed graphs

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

class DiGraph(object):
    '''
    A class to represent directed graphs.

    A *directed graph* is a couple (V,E) where V is a set of vertices and E
    is a set of edges (i.e., pairs of vertices). If (s,d) in E, then s and d
    are said *source* and *destination* of (s,d).
    '''

    def __init__(self, V=None, E=None):
        ''' Initialize a new DiGraph

        :param V: a collection of nodes
        :type V: a collection
        :param E: a collection of edges
        :type E: a collection
        '''

        self._next = dict()
        if V is not None:
            for v in V:
                self._next[v] = set()

        if E is not None:
            try:
                for src, dst in E:
                    if src in self._next:
                        self._next[src].add(dst)
                    else:
                        self._next[src] = set([dst])

                    if dst not in self._next:
                        self._next[dst] = set()
            except Exception:
                raise RuntimeError(('E = \'{}\' '.format(E)) +
                                   'must be a container of pairs ' +
                                   'of hashable objects.')

    def add_node(self, v):
        ''' Add a new node to a DiGraph

        :param self: the DiGraph object
        :type self: DiGraph
        :param v: a node
        '''
        if v in self._next:
            raise RuntimeError('v = \'{}\' '.format(v) +
                               'is already a node of this DiGraph')
        self._next[v] = set()

    def add_edge(self, src, dst):
        ''' Add a new edge to a DiGraph

        :param src: the source node of the edge
        :param dst: the destination node of the edge
        '''
        if src not in self._next:
            self._next[src] = set()
        else:
            if dst in self._next[src]:
                raise RuntimeError('({},{}) '.format(src, dst) +
                                   'is already an edge of this DiGraph')

        if dst not in self._next:
            self.add_node(dst)

        self._next[src].add(dst)

    def sources(self):
        ''' Return the sources of a DiGraph.

        The *sources* of a DiGraph G are the nodes that are sources of some
        edges in G itself.

        :returns: a generator of all the nodes that are sources of some edges
        :rtype: generator
        '''
        for src, dsts in self._next.items():
            if len(dsts) > 0:
                yield src

    def nodes(self):
        ''' Return the nodes of a DiGraph

        :returns: the list of the nodes of the DiGraph
        :rtype: list
        '''
        return self._next.keys()

    def next(self, src):
        ''' Return the next of a node

        Given a DiGraph :math:`(V,E)` and one of its node v, the
        *next* of :math:`v \in V` is the set of all those nodes :math:`v'` 
        that are destination of some edge :math:`(v,v') \in E`.

        :returns: the set of nodes :math:`\{v' | (v,v') \in E\}`
        :rtype: set
        '''
        if src not in self._next:
            raise RuntimeError('src = \'{}\' is not a node '.format(src) +
                               'of {}'.format(str(self)))

        return self._next[src]

    def edges(self):
        ''' Return the edges of a DiGraph

        :returns: a list of edges of the DiGraph
        :rtype: list
        '''

        return list(self.edges_iter())

    def edges_iter(self):
        ''' Return the edges of a DiGraph

        :returns: the generator of edges of the DiGraph
        :rtype: generator
        '''
        for src, dsts in self._next.items():
            for dst in dsts:
                yield (src, dst)

    def clone(self):
        ''' Clone a DiGraph

        :returns: a clone of the DiGraph
        :rtype: DiGraph
        '''
        nDG = DiGraph()

        nDG._next = dict()
        for src, dsts in self._next.items():
            nDG._next[src] = set(dsts)

        return nDG

    def __str__(self):
        ''' Return a string that represents a DiGraph

        :returns: a string that represents the DiGraph
        :rtype: str
        '''
        return '(V={}, E={})'.format(self.nodes(), self.edges())

    def get_subgraph(self, nodes):
        ''' Build the subgraph that respects a set of nodes

        :returns: the subgraph that respects :param nodes:
        :rtype: DiGraph
        '''

        V = set(nodes) & set(self.nodes())

        E = [(s, d) for (s, d) in self.edges_iter()
             if s in V and d in V]

        return DiGraph(V=V, E=E)

    def get_reversed_graph(self):
        ''' Build the reversed graph

        :returns: the reversed graph
        :rtype: DiGraph
        '''
        rE = [(d, s) for (s, d) in self.edges_iter()]

        return DiGraph(V=self.nodes(), E=rE)

    def get_reachable_set_from(self, nodes):
        ''' Compute the reachable set

        :param nodes: the set of nodes from which the reachability
                      should be evaluated
        :type nodes: a container of nodes
        :returns: the set of the reachable nodes
        :rtype: set
        '''
        queue = list(nodes)
        R = set(nodes)

        while queue:
            s = queue.pop()
            for d in self.next(s):
                if d not in R:
                    R.add(d)
                    queue.append(d)

        return R


def compute_SCCs(G):
    ''' Compute the strongly connected components of a DiGraph

    This method implements a non-recursive version of the
    Nuutila and Soisalon-Soinen's algorithm ([ns94]_) to compute the
    strongly connected components of a DiGraph.

    .. [ns94] E. Nuutila and E. Soisalon-Soinen. "On finding the strongly
       connected components in a directed graph.", Information Processing
       Letters 49(1): 9-14, (1994)

    :param G: the DiGraph object
    :type G: DiGraph
    :returns: a generator of the sets of nodes of the strongly connected
              components of the DiGraph
    :rtype: generator
    '''

    if not isinstance(G, DiGraph):
        raise TypeError('{} is not a DiGraph'.format(G))

    disc = dict()
    lowlink = dict()
    in_a_scc = set()
    scc_stack = []
    time = 0

    for s in G.nodes():
        if s not in disc:
            disc[s] = time
            lowlink[s] = time

            stack = [[s, iter(G.next(s))]]
            while stack:
                try:
                    w = next(stack[-1][1])

                    if w not in disc:
                        time = time+1
                        disc[w] = time
                        lowlink[w] = time

                        stack.append([w, iter(G.next(w))])

                except StopIteration:
                    [v, v_next_it] = stack.pop()

                    for w in G.next(v):
                        if w not in in_a_scc:
                            if disc[w] > disc[v]:
                                lowlink[v] = min([lowlink[v], lowlink[w]])
                            else:
                                lowlink[v] = min([lowlink[v], disc[w]])

                    if lowlink[v] == disc[v]:
                        in_a_scc.add(v)
                        scc = [v]
                        while scc_stack and disc[scc_stack[-1]] > disc[v]:
                            k = scc_stack.pop()
                            in_a_scc.add(k)
                            scc.append(k)
                        yield scc
                    else:
                        scc_stack.append(v)



"""A simple Python Model Checking package

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>

"""

__author__ = "Alberto Casagrande"
__copyright__ = "Copyright 2015-2020"
__credits__ = ["Alberto Casagrande"]
__license__ = "GPL"
__release__ = "1.3"
__subrelease__ = "3"
__version__ = __release__+"."+__subrelease__
__maintainer__ = "Alberto Casagrande"
__email__ = "acasagrande@units.it"
__status__ = "Development"

from .kripke import *
from .language import *

name = "pyModelChecking"
